# -*- coding: utf-8 -*-
from datetime import datetime

from sqlalchemy import Column, ForeignKey, Integer, PickleType, DateTime, \
        String, desc
from sqlalchemy.dialects.mysql import MEDIUMBLOB
from sqlalchemy.orm import relationship, deferred
from webob.multidict import MultiDict

from iktomi.utils import cached_property
from iktomi.cms.item_lock import ItemLock
from iktomi.cms.models.base import register_model

__all__ = ['EditLog']


def make_diff(field1, field2, changed=False):
    # XXX move to proper place
    if field1 is not None:
        label = field1.label or field1.name
        #field1.permissions = set('r')
        before = lambda: field1.widget.render()
    else:
        before = lambda: ''
    if field2 is not None:
        label = field2.label or field2.name
        name = field2.input_name
        #field2.permissions = set('r')
        after = lambda: field2.widget.render()
    else:
        after = lambda: ''
        name = ''

    return dict(label=label,
                name=name,
                before=before,
                after=after,
                changed=changed)


def _get_field_data(form, field):
    md = MultiDict()
    rv = field.from_python(form.python_data.get(field.name))
    field.set_raw_value(md, rv)
    return md



# XXX MySQL-specific type. How to resulve this?
class MediumPickleType(PickleType):

    impl = MEDIUMBLOB


@register_model('BaseModel')
def EditLogAdminUser(models):

    admin_id  = Column(ForeignKey(models.AdminUser.id, ondelete="CASCADE"),
                       primary_key=True)
    log_id  = Column(Integer, ForeignKey(models.EditLog.id, ondelete="CASCADE"),
                     primary_key=True)


@register_model('BaseModel')
def EditLog(models):

    id = Column(Integer, primary_key=True)
    stream_name = Column(String(100), nullable=False, default='')
    type = Column(String(50), nullable=False, default='edit')
    # object id can be string, so we use string here
    object_id = Column(String(100), nullable=True)
    global_id = Column(String(100), nullable=False, default='')
    edit_session = Column(String(50), nullable=False, default='')

    before = deferred(Column(MediumPickleType, default=list), group='data')
    after = deferred(Column(MediumPickleType, default=list), group='data')
    #diff = Column(Html(MediumText), nullable=True)

    creation_time = Column(DateTime, default=datetime.now, nullable=False)
    update_time = Column(DateTime, default=datetime.now, nullable=False)

    # there is a case when log item can be authored by multiple users:
    # if draft has been made by one user and than corrected and saved by other
    users = relationship(models.AdminUser,
                          secondary=models.EditLogAdminUser.__table__)

    @cached_property
    def data_changed(self):
        return self.before != self.after

    @classmethod
    def query_for_item(cls, db, item):
        ''''''
        global_id = ItemLock.item_global_id(item)
        return db.query(cls).filter_by(global_id=global_id)

    @classmethod
    def last_for_item(cls, db, stream_name, item, user, edit_session):
        ''''''
        # XXX filter by update_time?
        log = db.query(cls)\
                .filter(cls.users.contains(user))\
                .filter_by(stream_name=stream_name,
                           object_id=item.id)\
                .order_by(desc(cls.update_time))\
                .first()
        if log and log.edit_session == edit_session and log.type=='edit':
            return log

    __mapper_args__ = {'order_by': desc(update_time)}
    #Index('draft_index', stream_name, object_id)




# -*- coding: utf-8 -*-

import os
import logging
from webob.exc import HTTPMethodNotAllowed, HTTPBadRequest, HTTPNotFound
from iktomi import web
from iktomi.cms.stream_actions import PostAction
from iktomi.cms.forms.fields import AjaxImageField
from iktomi.cms.forms import convs
from iktomi.cms.stream_handlers import PrepareItemHandler, NoneIntConv
from iktomi.db.sqla.files import FileAttribute
from iktomi.db.sqla.images import ImageProperty
from iktomi.forms.files import check_file_path
from PIL import Image

logger = logging.getLogger(__file__)


class PrepareItemHandler(PrepareItemHandler):

    def retrieve_item(self, env, item):
        stream = self.action.stream
        model = stream.get_model(env)
        return env.db.query(model).filter_by(id=item).first()


class FileUploadHandler(web.WebHandler):

    def __init__(self, file_manager):
        self.file_manager = file_manager

    def get_length(self, request):
        length = int(request.environ.get('CONTENT_LENGTH', 0))
        if not length:
            from StringIO import StringIO
            from pprint import pprint
            sio = StringIO()
            sio.write('No content-length provided! \nHeaders:\n')
            pprint(request.headers.items(), sio)
            sio.write('\nEnviron:\n')
            pprint(request.environ, sio)
            sio.write('\nTrying to read directly from headers...')
            logger.warning(sio.getvalue())

            if request.headers.get('Content-Length'):
                length = int(request.headers.get('Content-Length'))
                logger.warning('YES! The Content-Length is in headers, '\
                               'but it can not be found in WSGI env!\n')
            elif 'content-length' in request.GET:
                length = int(request.GET['content-length'])
                logger.warning('Using a value from GET args: %s\n' % \
                                                                length)
            else:
                logger.warning('There is no Content-Length either in '\
                               'headers or in get args\n')
        return length

    def _get_file_manager(self, env):
        return self.file_manager

    def _save_file(self, env, data, length):
        file_manager = self._get_file_manager(env)
        return file_manager.create_transient(
                                   env.request.body_file_raw,\
                                   env.request.GET["file"],\
                                   length=length)

    def save_file(self, env, data, length):
        original_name = env.request.GET["file"]
        transient = self._save_file(env, data, length)
        file_manager = self._get_file_manager(env)
        return {
            "file": transient.name,
            'file_url': file_manager.get_transient_url(transient, env),
            'original_name': original_name,
            'mode':'transient'
            }

    def __call__(self, env, data):
        if env.request.method != "POST":
            raise HTTPMethodNotAllowed()

        length = self.get_length(env.request)
        if not length:
            return env.json({'status': 'failure',
                             'error': 'No Content-Length provided'})

        result = dict(status="ok")
        result.update(self.save_file(env, data, length))
        return env.json(result)


class StreamFileUploadHandler(PostAction, FileUploadHandler):

    item_lock = False
    for_item = True
    allowed_for_new = True
    display = False
    action = 'file_upload'

    @property
    def app(self):
        prefix = web.prefix('/<noneint:item>/'+self.action+'/<field_name>',
                            name=self.action,
                            convs={'noneint': NoneIntConv})
        prepare = PrepareItemHandler(self)
        return prefix | web.cases(
                  web.match() | prepare | self,
                  )

    def _get_file_manager(self, env):
        return env.db.find_file_manager(self.stream.get_model(env))


class StreamImageUploadHandler(StreamFileUploadHandler):

    force_autocrop = True
    action = 'image_upload'
    use_autocrop = True

    @property
    def app(self):
        prefix = web.prefix('/<noneint:item>/'+self.action+'/<field_name>',
                            name=self.action,
                            convs={'noneint': NoneIntConv})
        prepare = PrepareItemHandler(self)
        return prefix | web.cases(
                  web.match() | prepare | self,
                  web.match('/crop', 'crop') | prepare | self.crop,
                  )

    def _collect_related_fields(self, env, form_field, image,
                                original_name, ext):
        rel_images = []
        for name in dir(form_field.model):
            rel_field = getattr(form_field.model, name)
            if isinstance(rel_field, FileAttribute) and \
                    isinstance(rel_field.prop, ImageProperty) and \
                    rel_field.prop.fill_from == form_field.name:
                rel_form_field = form_field.name_parent.get_field(name)
                if rel_form_field is None:
                    continue
                if not self.force_autocrop and \
                        not getattr(rel_form_field.conv, 'autocrop', False):
                    continue
                if not self.use_autocrop:
                    rel_images.append({
                        "name": rel_form_field.input_name,
                        "mode":"empty",
                        "use_autocrop": self.use_autocrop,
                        })
                    rel_image = None
                else:
                    resizer = rel_field.prop.resize
                    target_size = rel_field.prop.image_sizes
                    transforms = resizer.transformations(image.size, target_size)
                    rel_image = resizer(image, target_size)
                    file_manager = self._get_file_manager(env)
                    rel_transient = file_manager.new_transient(ext)
                    rel_image.save(rel_transient.path)
                    rel_images.append({
                        "name": rel_form_field.input_name,
                        "file": rel_transient.name,
                        'file_url': file_manager.get_transient_url(rel_transient, env),
                        'fill_from': form_field.input_name,
                        'transformations': transforms,
                        'source_size': image.size,
                        'original_name': original_name,
                        'use_autocrop': self.use_autocrop,
                        'mode': 'empty'
                        })
                rel_images += self._collect_related_fields(
                                        env, rel_form_field,
                                        rel_image, original_name, ext)
        return rel_images

    def _get_form(self, env, data):
        item = data.item
        initial = {}
        if item is None:
            initial = data.filter_form.defaults()
            item = self.stream.get_model(env)(**initial)
        return self.stream.config.ItemForm.load_initial(env, item, initial=initial)

    def save_file(self, env, data, length):
        form = self._get_form(env, data)
        N_ = env.gettext
        #form.model = self.stream.get_model(env)
        field = form.get_field(data.field_name)
        if not isinstance(field, AjaxImageField) or \
                field.model is None:
            return FileUploadHandler.save_file(self, env, data, length)

        transient = self._save_file(env, data, length)
        try:
            image = Image.open(transient.path)
        except IOError:
            return {'status': 'failure',
                    'error': N_('Invalid image')}

        if image.size[0] * image.size[1] > getattr(env.cfg, 'MAX_IMAGE_SIZE', 5000*5000):
            return {'status': 'failure',
                    'error': N_('Image size exceeds the limit')}

        original_name = env.request.GET["file"]
        ext = os.path.splitext(original_name)[1]
        ext = ext or ('.' + (image.format or 'jpeg')).lower()

        rel_images = self._collect_related_fields(env, field, image,
                                                  original_name, ext)

        original_name = env.request.GET["file"]
        file_manager = self._get_file_manager(env)
        return {
            "file": transient.name,
            'file_url': file_manager.get_transient_url(transient, env),
            'original_name': original_name,
            'related_files': rel_images,
            'use_autocrop': self.use_autocrop,
            'mode':'transient'
            }

    def crop(self, env, data):
        item = data.item
        N_ = env.gettext

        if env.request.method != 'POST':
            raise HTTPMethodNotAllowed()

        fail = lambda msg: env.json(dict(status='failure', error=msg))

        form = self._get_form(env, data)
        #form.model = self.stream.get_model(env)
        form_field = form.get_field(data.field_name)
        if not isinstance(form_field, AjaxImageField) or \
                form_field.model is None:
            raise HTTPNotFound()

        model_field = getattr(form_field.model, form_field.name)

        if not model_field.prop.fill_from:
            raise HTTPNotFound()

        if env.request.POST.get('mode') == 'existing':
            fill_from = model_field.prop.fill_from
            source = getattr(item, fill_from)
            if source is None:
                raise HTTPNotFound
            source = source.path
        elif env.request.POST.get('mode') == 'transient':
            transient_name = env.request.POST.get('transient_name')
            try:
                transient_name = check_file_path(None, transient_name)
            except convs.ValidationError:
                raise HTTPBadRequest()
            file_manager = self._get_file_manager(env)
            source = file_manager.get_transient(transient_name).path
        else:
            return fail(N_('Invalid mode'))

        _, original_name = os.path.split(source)

        try:
            image = Image.open(source)
        except IOError:
            return fail(N_('Invalid image'))

        box = ()
        for f in  ['left', 'top', 'right', 'bottom']:
            try:
                box += (int(env.request.POST.get(f, '')), )
            except ValueError:
                return fail(N_('Invalid coordinates'))

        image = image.crop(box)
        ext = os.path.splitext(source)[1]
        ext = ext or ('.' + (image.format or 'jpeg')).lower()
        file_manager = self._get_file_manager(env)
        transient = file_manager.new_transient(ext)
        image.save(transient.path, quality=100)

        rel_images = self._collect_related_fields(
                                    env, form_field, image, original_name, ext)

        return env.json({
            'status': 'ok',
            'file': transient.name,
            'file_url': file_manager.get_transient_url(transient, env),
            'original_name': original_name,
            'related_files': rel_images,
            'use_autocrop' : self.use_autocrop,
            'mode' : 'transient'
            })



# -*- coding: utf-8 -*-
from sqlalchemy import Column, Integer, DateTime, Boolean
from sqlalchemy.sql.expression import text
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import object_session
from sqlalchemy.orm.util import identity_key

from iktomi.unstable.db.sqla.replication import replicate, replicate_attributes
from iktomi.unstable.db.sqla.public_query import PublicQuery
from iktomi.utils import cached_property, cached_class_property
from iktomi.cms.item_lock import ItemLock

from datetime import datetime


class AdminPublicQuery(PublicQuery):

    property_name = 'existing'


class _WithState(object):

    ABSENT = 0
    PRIVATE = 1
    PUBLIC = 2
    DELETED = 3

    @declared_attr
    def created_dt(self):
        return Column(DateTime, nullable=False, default=datetime.now)

    @declared_attr
    def updated_dt(self):
        return Column(DateTime, nullable=False, default=datetime.now,
                      onupdate=datetime.now)

    @hybrid_property
    def public(self):
        return self.state==self.PUBLIC

    # public condition for admin site
    @hybrid_property
    def existing(self):
        return self.state in [self.PUBLIC, self.PRIVATE]

    @existing.expression
    def existing(cls):
        return cls.state.in_([cls.PUBLIC, cls.PRIVATE])


class WithState(_WithState):

    @declared_attr
    def state(self):
        return Column(Integer, nullable=True, default=self.ABSENT)


class _AdminWithStateMixIn(object):

    def publish(self):
        self.state = self.PUBLIC
        _AdminReplicated.publish(self)

    def unpublish(self):
        '''Make the object invisible on front site.'''
        self._front_item.state = self.state = self.PRIVATE

    def delete(self):
        '''Mark the object as deleted (make it invisible on front and move to
        trash on admin site).'''
        self._front_item.state = self.state = self.DELETED


class _FrontOnlyWithState(_WithState):

    def publish(self):
        self.state = self.PUBLIC

    def unpublish(self):
        '''Make the object invisible on front site.'''
        self.state = self.PRIVATE

    def delete(self):
        self.state = self.DELETED


class FrontOnlyWithState(_FrontOnlyWithState, WithState):
    '''Class with publishing and without replication'''
    pass


class _AdminWithState(_AdminWithStateMixIn, _WithState):
    pass


class AdminWithState(_AdminWithStateMixIn, WithState):

    pass


def _get_model_name(item):
    modelname = item.__class__.__name__
    # XXX item.models is not an interface
    for lang in item._iktomi_langs:
        if modelname.endswith(lang.title()):
            return modelname[:-len(lang)]


class WithLanguage(object):

    def _item_version(self, version, lang=None):
        lang = lang or self.models.lang
        if not lang in self._iktomi_langs:
            return None
        # XXX hacky
        models = getattr(AdminReplicated, version)
        models = getattr(models, lang)
        modelname = _get_model_name(self)
        model = getattr(models, modelname)
        db = object_session(self)
        ident = identity_key(instance=self)[1]
        assert ident is not None
        return db.query(model).get(ident)


class AdminWithLanguage(WithLanguage):

    def _create_version(self, cls):
        return cls(id=self.id)

    def _create_versions(self):
        # be careful! makes flush!
        db = object_session(self)
        modelname = _get_model_name(self)

        main_lang = self.models.main_lang
        if not main_lang in self._iktomi_langs:
            main_lang = self._iktomi_langs[0]

        # The first language is default. It is used as id autoincrement
        if self.models.lang != main_lang and self.id is None:
            model = getattr(self.models,
                            modelname + main_lang.title())
            ru = self._create_version(model)
            # Flush ru model first to get autoincrement id.
            # Do not flush english model yet, but return it to pending
            # state after the flush
            db.add(ru)
            db.expunge(self)
            db.flush()
            self.id = ru.id
            db.add(self)
            db.flush() # XXX is this needed?
        elif self.id is None:
            db.flush()

        for lang in self._iktomi_langs:
            # create all en/ru admin/front versions
            if lang == self.models.lang:
                item = self
            else:
                model = getattr(self.models, modelname + lang.title())
                item = db.query(model).get(self.id)
                if item is None:
                    item = self._create_version(model)
                    db.add(item)

            if item._front_item is None:
                # Front item will be created right now,
                # `_front_item` is cached_property,
                # so we should invalidate it's value
                del item.__dict__['_front_item']
                # XXX it is better to do this automatically on before_insert or
                #     after_insert
                item._create_front_object()
            # flush changes for each language separately
            # to force queries in right order also on the front
            db.flush()

        if hasattr(self, 'state'): # XXX or isinstance check?
            if self.state is None or self.state in (self.ABSENT, self.DELETED):
                self.state = self.PRIVATE


# ==============   Repicated models


class ReplicatedVersions(object):

    def _item_version(self, version):
        # XXX hacky
        models = getattr(AdminReplicated, version)
        model = getattr(models, self.__class__.__name__)
        db = object_session(self)
        ident = identity_key(instance=self)[1]
        assert ident is not None
        return db.query(model).get(ident)


class _FrontReplicated(object):


    @cached_class_property
    def _admin_model(cls):
        # XXX
        return getattr(AdminReplicated.admin, cls.__name__)

    @cached_property
    def _admin_item(self):
        db = object_session(self)
        assert db is not None
        ident = identity_key(instance=self)[1]
        assert ident is not None
        return db.query(self._admin_model).get(ident)

    @cached_property
    def _other_version(self):
        return self._admin_item

    def item_global_id(self):
        gid = ItemLock.item_global_id(self, view_in_obj=False)
        assert gid.startswith(self.__class__.__module__)
        return gid.replace(self.__class__.__module__, 
                           self._admin_model.__class__.__module__, 1)
        #return ItemLock.item_global_id(self._admin_item)


class FrontReplicated(ReplicatedVersions, _FrontReplicated):

    @cached_property
    def has_unpublished_changes(self):
        return self._admin_item.has_unpublished_changes


class _AdminReplicated(object):
    '''Model that is replicated (published) to front. You always have two
    versions: current and published.
    Don't use it for secondary table models, models of in-place editable relations
    and other models that are not replicated directly
    (use base model without mixins instead).'''

    @cached_class_property
    def front(cls):
        # XXX
        from models import front
        return front

    @cached_class_property
    def admin(cls):
        # XXX cycled imports
        from models import admin
        return admin

    @cached_class_property
    def _front_model(cls):
        return getattr(AdminReplicated.front, cls.__name__)

    @cached_property
    def _front_item(self):
        db = object_session(self)
        assert db is not None
        ident = identity_key(instance=self)[1]
        assert ident is not None
        return db.query(self._front_model).get(ident)

    @cached_property
    def _other_version(self):
        return self._front_item

    def _create_versions(self):
        # be careful! makes flush!
        if self._front_item is None:
            db = object_session(self)
            db.flush()
            # Front item will be created right now,
            # `_front_item` is cached_property,
            # so we should invalidate it's value
            del self.__dict__['_front_item']
            # XXX it is better to do this automatically on before_insert or
            #     after_insert
            self._create_front_object()
            db.flush()
        # the item is created, we set PRIVATE state as default
        # XXX hasattr looks hacky
        if hasattr(self, 'state'):
            if self.state is None or self.state in (self.ABSENT, self.DELETED):
                self.state = self.PRIVATE

    def _create_front_object(self):
        replicate(self, self._front_model)

    def _copy_to_front(self):
        replicate(self, self._front_model)

    def _copy_from_front(self):
        replicate_attributes(self._front_item, self)

    def publish(self):
        '''Publish current version: copy to front and change auxiliary fields
        appropriately.'''
        self.has_unpublished_changes = False
        self._copy_to_front()

    def revert_to_published(self):
        '''Revert to published version (undo all changes after the object was
        last published).'''
        self._copy_from_front()
        self.has_unpublished_changes = False


class AdminReplicated(ReplicatedVersions, _AdminReplicated):

    @declared_attr
    def has_unpublished_changes(self):
        '''Was the object updated after publishing? Does it differ from
        published version?'''
        # XXX why default was set to True?
        return Column(Boolean, nullable=False, default=False,
                      server_default='0')#, onupdate=True)




# -*- coding: utf-8 -*-
import inspect
import json
from datetime import datetime
from webob.exc import HTTPNotFound, HTTPForbidden, HTTPOk
from webob.multidict import MultiDict
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import class_mapper, RelationshipProperty

from iktomi.utils.deprecation import deprecated
from iktomi.utils import cached_property
from iktomi import web
from iktomi.utils.paginator import ModelPaginator, FancyPageRange
from iktomi.web.url_converters import Integer as IntegerConv, \
                        Converter as BaseConv, ConvertError

from .item_lock import ItemLockData
from .item_lock import ModelLockError, ItemLock
from .stream_actions import StreamAction
from .flashmessages import flash


def see_other(location):
    return web.Response(json.dumps({'location': location}),
                        content_type="application/json")

def insure_is_xhr(env):
    if (env.request.method in ('GET', 'HEAD') and \
           not env.request.is_xhr and \
           not '__ajax' in env.request.GET):
        raise HTTPOk(body=env.render_to_string('layout.html', {}))


class NoneIntConv(IntegerConv):
    name = 'noneint'
    regex = BaseConv.regex

    @deprecated('Use stream property id_none_converter instead')
    def __init(self, *args, **kwargs):
        super(NoneIntConv, self).__init__(*args, **kwargs)

    def to_python(self, value, **kwargs):
        if value == '+':
            return None
        try:
            value = int(value)
        except ValueError:
            raise ConvertError(self, value)
        else:
            return value

    def to_url(self, value):
        if value is None:
            return '+'
        return str(value)


class StreamListHandler(StreamAction):

    for_item = False
    item_lock = False
    display = False

    @cached_property
    def template_name(self):
        return self.stream.stream_template_name

    @cached_property
    def base_template(self):
        return getattr(self.stream, 'list_base_template', 'stream_base.html')

    @property
    def app(self):
        return web.match() | self

    def list_handler(self, env, data):
        #insure_is_xhr(env)

        stream = self.stream
        stream.insure_has_permission(env, 'x')
        read_allowed = stream.has_permission(env, 'r')
        template_name = stream.row_template_name
        template_data = stream.template_data
        no_layout = ('__no_layout' in env.request.GET)

        def item_row(item, list_fields=None, url='#', row_cls='', **kw):
            return env.render_to_string(template_name, dict(template_data,
                                             item=item,
                                             list_fields=list_fields,
                                             url=url,
                                             row_cls=row_cls,
                                             read_allowed=read_allowed,
                                             stream=stream,
                                             **kw))

        data = dict(self.prepare_data(env, data),
                    is_popup=('__popup' in env.request.GET),
                    no_layout=no_layout,
                    menu=stream.module_name,
                    item_row=item_row,
                    base_template=self.base_template)

        return env.render_to_response(self.template_name, data)
    __call__ = list_handler

    def prepare_data(self, env, data):
        '''View for index page.'''
        stream = self.stream
        request = env.request

        # XXX Check permissions and return 403 if no access.
        # XXX Handle POST to edit/delete here.
        query = stream.item_query(env)
        filter_form = stream.get_filter_form(env)

        # Note: errors are displayed, but ignored in code.
        filter_form.accept(request.GET)
        filter_data = filter_form.get_data()
        query = filter_form.filter(query)

        query = stream.order(query)

        paginator = ModelPaginator(request, query,
                                   impl=FancyPageRange(),
                                   limit=getattr(stream.config, 'limit', None),
                                   url=stream.url_for(env).qs_set(filter_data))

        def item_url(item=None):
            item_id = item.id if item is not None else None
            return stream.url_for(env, 'item', item=item_id).qs_set(
                                                                filter_data)

        try:
            paginator.items = stream.config.modify_items(paginator.items)
        except AttributeError:
            pass


        result = dict(stream.template_data,
                      paginator=paginator,
                      stream=stream,
                      item_url=item_url,
                      list_fields=stream.list_fields,
                      title=stream.title,
                      filter_form=filter_form,
                      allow_add=stream.has_permission(env, 'c'),
                      live_search=stream.live_search,
                      repr=repr)
        result.update(self.list_form_data(env, paginator, filter_data))
        result = stream.process_list_template_data(env, result)
        return result

    def list_form_data(self, env, paginator, filter_data):
        use_list_form = self.stream.ListItemForm and \
                self.stream.list_edit_action.save_allowed(env) and \
                (self.stream.list_edit_action.use_with_filters or not filter_data)

        if use_list_form:
            return {'list_item_form':self.stream.ListItemForm(
                                                 env, items=paginator.items)}
        return {}



class PrepareItemHandler(web.WebHandler):
    """ Helper handler to fetch item by id field.
    `data` in handler should have `item` attr containts item id.
    """
    # XXX: Add support for fetching by pk, not only `id` field
    #       (use query.get(pk_val) )

    def __init__(self, action):
        self.action = action

    def retrieve_item(self, env, item):
        return self.action.stream.item_query(env).filter_by(id=item).first()

    def take_lock(self, env, data):
        if self.action.item_lock:
            stream = self.action.stream
            data.item_lock = ItemLockData.for_item(env, stream, data.item,
                                                   data.filter_form)
            data.lock_message = data.item_lock.message
            data.edit_session = data.item_lock.edit_session
        else:
            data.item_lock = None
            data.lock_message = data.edit_session = ''

    def prepare_item_handler(self, env, data):
        '''Item actions dispatcher'''
        #if self.action.xhr:
        #    insure_is_xhr(env)

        stream = self.action.stream
        stream.insure_has_permission(env, 'r')

        data.filter_form = stream.get_filter_form(env)
        # Note: errors are displayed, but ignored in code.
        data.filter_form.accept(env.request.GET)

        if data.item is not None:
            data.item = self.retrieve_item(env, data.item)
            if data.item is None:
                raise HTTPNotFound
        elif not self.action.allowed_for_new:
            flash(env, u'Действие «%s» недоступно для нового объекта' %
                  (self.action.title,),
                  'failure')
            raise HTTPNotFound

        self.take_lock(env, data)
        return self.next_handler(env, data)
    __call__ = prepare_item_handler


class EditItemHandler(StreamAction):

    action = 'item'
    item_lock = True
    allowed_for_new = True
    display = False
    title = u'Редактировать/создать'
    PrepareItemHandler = PrepareItemHandler

    @property
    def app_prefix(self):
        return web.prefix('/<idconv:item>', name='item',
                          convs={'idconv': self.stream.id_none_converter})
    @property
    def app(self):
        # Be careful, put prepare handler only after match! Otherwise it brakes
        # file upload (yes, its not obvious!)
        prepare =  self.PrepareItemHandler(self)
        return self.app_prefix | web.cases(
                web.match('', '') | prepare | self,
                web.match('/autosave', 'autosave') | \
                        web.method('POST', strict=True) | prepare |self.autosave
            )

    def create_allowed(self, env):
        return self.stream.has_permission(env, 'c')

    def save_allowed(self, env, item=None):
        return self.stream.has_permission(env, 'w')

    def delete_allowed(self, env, item=None):
        return self.stream.has_permission(env, 'd')

    def get_item_template(self, env, item):
        return self.stream.item_template_name

    def _clean_item_data(self, stream, env, item):
        form_cls = stream.config.ItemForm
        form = form_cls.load_initial(env, item, initial={}, permissions='r', for_diff=True)
        raw_data = form.raw_data.items()
        unic = lambda s: (s if type(s) in (unicode, bytes) else unicode(s))
        return [(unic(k), unic(v)) for k, v in raw_data]

    def get_item_form(self, stream, env, item, initial, draft=None):
        save_allowed = self.create_allowed(env) \
                if (item is None or item.id is None) else \
                self.save_allowed(env, item)
        form_kw = {}
        if not save_allowed:
            form_kw['permissions'] = 'r'

        form_cls = stream.config.ItemForm
        form = form_cls.load_initial(env, item, initial=initial, **form_kw)
        form.model = self.stream.get_model(env)
        form.draft = draft
        if draft is not None:
            raw_data = MultiDict(draft.data)
            form.accept(raw_data) # XXX
        return form

    def save_item(self, env, filter_form, form, item, draft, autosave):
        form.update_instance(item)

        if item not in env.db:
            env.db.add(item)
        if draft is not None:
            env.db.delete(draft)

        self.stream.commit_item_transaction(env, item, silent=autosave)
        if hasattr(self, 'post_create'):
            self.post_create(item)

        item_url = self.stream.url_for(env, 'item', item=item.id)\
                              .qs_set(filter_form.get_data())
        autosave_url = self.stream.url_for(env, 'item.autosave', item=item.id)\
                                  .qs_set(filter_form.get_data())
        return item, item_url, autosave_url

    def get_log_item(self, env, data, item):
        EditLog = env.edit_log_model
        stream = self.stream
        log = EditLog.last_for_item(
                env.db, stream.uid(env), item, 
                env.user, data.edit_session)
        if log is None:
            before = self._clean_item_data(stream, env, item)
            if (item.id is None
                    or hasattr(item, 'existing') and not item.existing):
                action_type = 'create'
            else:
                action_type = 'edit'
            log = EditLog(stream_name=stream.uid(env),
                          type=action_type,
                          object_id=item.id,
                          global_id=ItemLock.item_global_id(item),
                          edit_session=data.edit_session,
                          users=[env.user],
                          before=before)
        return log

    def save_log_item(self, env, data, log, item):
        env.db.commit()
        # do not write form.raw_data directly, because it can be
        # different for same data stored in db.
        # Assume form raw data generated from item as canonical
        # representation
        log.after = self._clean_item_data(self.stream, env, item)
        if log.data_changed:
            log.object_id = item.id
            log.global_id = ItemLock.item_global_id(item)
            log.update_time = datetime.now()
            env.db.add(log)
            env.db.commit()

    def edit_item_handler(self, env, data):
        '''View for item page.'''

        item, lock_message, filter_form = \
            data.item, data.lock_message, data.filter_form
        stream = self.stream
        request = env.request

        initial = filter_form.defaults()
        stream_url = stream.url_for(env).qs_set(filter_form.get_data())
        create_allowed = save_allowed = self.create_allowed(env)
        delete_allowed = False
        success = False

        if item is None:
            if not create_allowed:
                raise HTTPForbidden
            # We must pass initial data to allow creation of models with
            # inheritance.
            item = stream.get_model(env)(**initial)
            # Don't add it to session since we don't know yet whether it should
            # be saved (there can be errors in form).
            if data.item_lock:
                data.item_lock.item = item # XXX
        else:
            save_allowed = self.save_allowed(env, item)
            delete_allowed = self.delete_allowed(env, item)


        autosave_allowed = save_allowed and \
                           getattr(env, 'draft_form_model', None) and \
                           stream.autosave
        autosave = autosave_allowed and getattr(data, 'autosave', False)
        if autosave_allowed:
            DraftForm = env.draft_form_model
            draft = DraftForm.get_for_item(env.db, stream.uid(env),
                                           item, env.user)
            has_draft = bool(draft)
        elif getattr(data, 'autosave', False):
            raise HTTPForbidden
        else:
            draft = None
            has_draft = False

        if item.id is None \
                and not getattr(data, 'autosave', False) \
                and 'force_draft' not in env.request.GET:
            draft = None

        form = self.get_item_form(stream, env, item, initial, draft)
        EditLog = getattr(env, 'edit_log_model', None)
        log_enabled = (EditLog is not None and
                       getattr(env, 'version', 'admin') == 'admin' and
                       stream.edit_log_action is not None)

        if request.method == 'POST':
            if not save_allowed:
                raise HTTPForbidden

            log = None
            if log_enabled:
                log = self.get_log_item(env, data, item)

            accepted = form.accept(request.POST)
            if accepted and not lock_message:
                need_lock = item.id is None and self.item_lock and autosave
                item, item_url, autosave_url = \
                        self.save_item(env, filter_form, form,
                                       item, draft, autosave)

                if log is not None:
                    self.save_log_item(env, data, log, item)

                result = {'success': True,
                          'item_id': item.id,
                          'autosave_url': autosave_url,
                          'item_url': item_url}
                if need_lock:
                    try:
                        gid = ItemLock.item_global_id(item)
                        lock = env.item_lock.create(gid)
                    except ModelLockError, e:
                        flash(env, unicode(e))
                    else:
                        result = dict(result,
                            edit_session=lock,
                            status='captured',
                            global_id=gid)
                return env.json(result)
            elif lock_message and autosave:
                stream.rollback_due_lock_lost(env, item)
                return env.json({'success': False,
                                 'error': 'item_lock',
                                 'lock_message': lock_message})
            elif lock_message:
                stream.rollback_due_lock_lost(env, item)
            elif autosave:
                stream.rollback_due_form_errors(env, item, silent=True)
                if draft is None:
                    draft = DraftForm(stream_name=stream.uid(env),
                                      object_id=item.id)
                    env.db.add(draft)
                draft.data = form.raw_data.items()

                if not env.user in draft.users:
                    draft.users.append(env.user)
                draft.update_time = datetime.now()
                env.db.commit()
                return env.json({'success': False,
                                 'error': 'draft',
                                 'draft_id': draft.id,
                                 'errors': form.errors,
                                 })
            else:
                stream.rollback_due_form_errors(env, item, silent=autosave)

        template_data = dict(filter_form=filter_form,
                             success=success,
                             form=form,
                             roles=env.user.roles,
                             item=item,
                             draft=draft,
                             has_draft=has_draft,
                             stream=stream,
                             stream_title=stream.config.title,
                             title=unicode(item),
                             log_enabled=log_enabled,
                             submit_url=stream.url_for(env, 'item',
                                                    item=item.id).qs_set(
                                                        filter_form.get_data()),

                             menu=stream.module_name,
                             stream_url=stream_url,

                             item_lock=data.item_lock,

                             actions=[x for x in stream.actions 
                                      if x.for_item and x.is_visible(env, item)],
                             item_buttons=stream.buttons,

                             list_allowed=self.stream.has_permission(env, 'x'),
                             autosave_allowed=autosave_allowed,
                             create_allowed=create_allowed,
                             save_allowed=save_allowed,
                             delete_allowed=delete_allowed,

                             is_popup=('__popup' in request.GET))

        template_data = stream.process_item_template_data(env, template_data)
        template_data = self.process_item_template_data(env, template_data)

        return env.render_to_response(self.get_item_template(env, item),
                                      template_data)
    __call__ = edit_item_handler


    def autosave(self, env, data):
        data.autosave = True
        return self(env, data)

    def process_item_template_data(self, env, template_data):
        '''Preprocessor for template variables.
           Can be overriden by descedant classes.'''
        return template_data


def _get_all_classes(model):
    metadata = model.metadata
    models = [x for x in inspect.getmro(model)
              if hasattr(x, 'metadata') and x.metadata is metadata]
    set_models = set(models)
    base_models = set(models)
    for m in models:
        if set(m.__bases__) & set_models:
            base_models.remove(m)

    _seen = set()
    def _iter_subclasses(cls):
        for sub in cls.__subclasses__():
            if sub not in _seen:
                _seen.add(sub)
                yield sub
                for sub in _iter_subclasses(sub):
                    yield sub

    models = set()
    for m in base_models:
        models |= set([x for x in _iter_subclasses(m)
                       if not x.__dict__.get('__abstract__', False)])
    return models

def _get_referers(db, item):
    '''Returns a dictionary mapping referer model class to query of all
    objects of this class refering to current object.'''
    cls = type(item)
    result = {}
    for other_class in _get_all_classes(cls):
        queries = {}
        for prop in class_mapper(other_class).iterate_properties:
            if not (isinstance(prop, RelationshipProperty) and \
                    issubclass(cls, prop.mapper.class_)):
                continue
            query = db.query(prop.parent)
            comp = prop.comparator
            if prop.uselist:
                query = query.filter(comp.contains(item))
            else:
                query = query.filter(comp==item)
            count = query.count()
            if count:
                queries[prop] = (count, query)
        if queries:
            result[other_class] = queries
    return result


class _ReferrersAction(StreamAction):

    def _list_referers(self, env, obj, limit=50, exclude=None):
        if exclude is None:
            exclude = set()
        result = {}
        for cls, props in _get_referers(env.db, obj).items():
            for prop, (count, query) in props.items():
                left = limit - len(result)
                if left <= 0:
                    return result
                indirect = []
                for ref in query[:left]:
                    if ref == obj or ref in exclude or ref in result:
                        continue
                    # XXX this is not optimal way to get edit url
                    for stream in env.streams.values():
                        url = stream.get_edit_url(env, ref)
                        if url is not None:
                            result[ref] = url
                            break
                    else:
                        indirect.append(ref)
                assert obj not in result
                for ref in indirect:
                    left = limit - len(result)
                    if left <= 0:
                        return result
                    indirect_exclude = exclude | set([obj]) | set(result) | \
                        set(indirect)
                    indirect_referers = self._list_referers(
                        env, ref, limit=left,
                        exclude=indirect_exclude)
                    # XXX The following creates false warnings for DocLink
                    # objects that are "private" to current doc.
                    #if not (exclude or indirect_referers):
                    #    # We have to insure that there is a link for each top
                    #    # level referer (that's why exclude is checked).
                    #    logger.warning("Can't find editable object for %s",
                    #                   ref)
                    result.update(indirect_referers)
                    assert obj not in result
        return result


class DeleteItemHandler(_ReferrersAction):

    action = 'delete'
    cls = 'delete'
    title = u'Удалить'
    PrepareItemHandler = PrepareItemHandler

    @property
    def app(self):
        return web.match('/<idconv:item>/delete', 'delete',
                         convs={'idconv': self.stream.id_converter}) | \
            self.PrepareItemHandler(self) | self

    def is_available(self, env, item):
        return StreamAction.is_available(self, env, item) and \
                self.stream.has_permission(env, 'd')

    def clear_tray(self, env, item):
        ObjectTray = env.object_tray_model
        stream_name = env.stream.uid(env, version=False)
        tray_objects = env.db.query(ObjectTray)\
                             .filter_by(object_id=item.id)\
                             .filter_by(stream_name=stream_name)\
                             .all()
        for obj in tray_objects:
            env.db.delete(obj)

    def delete_item_handler(self, env, data):
        #insure_is_xhr(env)
        item, edit_session, lock_message, filter_form = \
            data.item, data.edit_session, data.lock_message, data.filter_form
        stream = self.stream

        self.insure_is_available(env, item)

        stream_url = stream.url_for(env).qs_set(filter_form.get_data())
        item_url = stream.url_for(env, 'item', item=item.id).qs_set(
                                   filter_form.get_data())
        delete_url = stream.url_for(env, 'delete', item=item.id)\
                           .qs_set(filter_form.get_data())
        if env.request.method == 'POST':
            env.db.delete(item)
            log = self.stream.create_log_entry(env, data.item, 'delete')
            if log is not None:
                env.db.add(log)
            if getattr(env, 'object_tray_model', None):
                self.clear_tray(env, item)
            try:
                env.db.commit()
            except IntegrityError:
                env.db.rollback()
                flash(env, u'Невозможно удалить объект (%s) пока на него'\
                      u' ссылаются другие объекты' % (item,),
                      'failure')
                return env.json({'result': 'failure'})
            return env.json({'result': 'success',
                             'location': stream_url})
        data = dict(item=item,
                    item_url=item_url,
                    form_url=delete_url,
                    referers=self._list_referers(env, item),
                    title=u'Удаление объекта «%s»' % item,
                    stream=stream,
                    stream_url=stream_url,
                    menu=stream.module_name)
        return env.render_to_response('delete', data)
    __call__ = delete_item_handler


class GetReferrersHandler(_ReferrersAction):

    mode = 'popup'
    item_lock = False
    accepts_item_form = False
    display = False
    action = 'referrers'
    title = u'Связанные объекты'
    PrepareItemHandler = PrepareItemHandler

    @property
    def app(self):
        return web.match('/<idconv:item>/{}'.format(self.action),
                         self.action,
                         convs={'idconv': self.stream.id_converter}) | \
            self.PrepareItemHandler(self) | self

    def referrers_handler(self, env, data):
        insure_is_xhr(env)
        data = dict(item=data.item,
                    referers=self._list_referers(env, data.item))
        return env.render_to_response('referrers', data)
    __call__ = referrers_handler




# -*- coding: utf-8 -*-
from iktomi.utils import cached_property
from iktomi import web
from .stream_handlers import EditItemHandler, PrepareItemHandler
from .stream import Stream

class PrepareLonerHandler(PrepareItemHandler):

    def retrieve_item(self, env, item):
        return self.action.stream.retrieve_item(env, item)

    def prepare_item_handler(self, env, data):
        '''Item actions dispatcher'''
        #insure_is_xhr(env)

        stream = self.action.stream
        stream.insure_has_permission(env, 'r')

        data.filter_form = stream.get_filter_form(env) # XXX

        data.item = self.retrieve_item(env, None)
        self.take_lock(env, data)
        return self.next_handler(env, data)
    __call__ = prepare_item_handler


class LonerHandler(EditItemHandler):

    PrepareItemHandler = PrepareLonerHandler

    @property
    def app(self):
        prepare = self.PrepareItemHandler(self)
        return web.cases(
                web.match('', '') | prepare | self,
                web.match('/autosave', 'autosave') | \
                        web.method('POST', strict=True) | prepare | self.autosave
            )


class Loner(Stream):

    core_actions = [LonerHandler()]

    lock_back_title = u'Вернуться на главную'
    lock_back_help = u'Вернуться на главную страницу'

    #def uid(self, env, version=True):
    #    # Attention! Be careful!
    #    # Do not change format of uid unless you are sure it will not 
    #    # brake tray views, where stream_name and language are parsed out
    #    # from the uid
    #    return 'loners.' + self.module_name

    @cached_property
    def extra_filters(self):
        return getattr(self.config, 'model_filters', {})

    def url_for(self, env, name=None, **kwargs):
        kwargs.pop('item', 0)
        if name and name[:5] in ['item.', 'item']:
            name = name[5:] # XXX
        name = name and '%s.%s' % (self.module_name, name) or self.module_name
        return env.url_for(name, **kwargs)

    @cached_property
    def perms(self):
        p = getattr(self.config, 'permissions', {})
        p.setdefault('wheel', 'rw')
        return p

    def lock_back_url(self, env, item, filter_form):
        return env.root.index

    def retrieve_item(self, env, item):
        Model = self.get_model(env)
        extra_filters = self.extra_filters

        item = env.db.query(Model)\
                     .filter_by(**extra_filters).scalar()
        if item is None:
            item = Model(**extra_filters)
        return item





# -*- coding: utf-8 -*-
from cStringIO import StringIO
from jinja2 import Markup
from iktomi.forms.form import Form
from iktomi.forms.files import FileFieldSet, FileFieldSetConv
from iktomi.db.files import PersistentFile
from iktomi.db.sqla.files import FileAttribute
from iktomi.db.sqla.images import ImageProperty
from iktomi.unstable.utils.image_resizers import ResizeFit
from iktomi.utils import cached_property
from iktomi.cms.forms import convs, widgets
from PIL import Image


class AjaxFileField(FileFieldSet):

    widget = widgets.FieldSetWidget(template='widgets/ajax_fileinput')

    @property
    def upload_url(self):
        env = self.env
        # XXX looks like a hack
        if any(x for x in env.stream.actions if x.action=="file_upload"):
            return env.stream.url_for(env, 'file_upload',
                                      item=self.form.item.id,
                                      field_name=self.input_name)
        return env.root.load_tmp_file

    def __init__(self, *args, **kwargs):
        required = kwargs.pop('required', None)
        if required is not None:
            conv = kwargs.get('conv', self.conv)
            kwargs['conv'] = conv(required=required)
        FileFieldSet.__init__(self, *args, **kwargs)

    def set_raw_value(self, raw_data, value):
        FileFieldSet.set_raw_value(self, raw_data, value)
        # XXX hack!
        # XXX removing for_diff check breaks history logging, adding it breaks changed 
        #     fields indication. Workaround is needed
        if getattr(self.form, 'for_diff', False) and self.clean_value:
            raw_data[self.prefix+'path'] = self.clean_value.name

    def get_diff(field1, field2):
        path1 = field1.form.raw_data.get(field1.prefix+'path')
        path2 = field2.form.raw_data.get(field2.prefix+'path')
        if path1 != path2:
            return dict(label=field2.label,
                        name=field2.input_name,
                        before=lambda: path1,
                        after=lambda: path2,
                        changed=True)


class ImageFieldSetConv(FileFieldSetConv):

    autocrop = False

    def to_python(self, value):
        if value["mode"] == "existing" and self.autocrop:
            source = self.field.form.get_field(self.field.fill_from)
            if not isinstance(source.clean_value, PersistentFile):
                value = dict(mode="empty")
        return FileFieldSetConv.to_python(self, value)


class AjaxImageField(AjaxFileField):

    widget = widgets.FieldSetWidget(template='widgets/ajax_imageinput')
    #: used when file is uploaded
    thumb_size = None
    model_field = None

    # Show the thumb or not. Works only with image option on
    show_thumbnail = True
    show_size = True
    #: use js pre-upload thumbnail generation by canvas
    #canvas_thumb_preview = False
    conv = ImageFieldSetConv()
    retina = False
    crop = True

    def __init__(self, *args, **kwargs):
        kwargs['_label'] = kwargs.pop('label', kwargs.get('_label'))
        AjaxFileField.__init__(self, *args, **kwargs)

    @property
    def upload_url(self):
        env = self.env
        # XXX looks like a hack
        if any(x for x in env.stream.actions if x.action=="image_upload"):
            return env.stream.url_for(env, 'image_upload',
                                      item=self.form.item.id,
                                      field_name=self.input_name)
        return env.root.load_tmp_file

    @property
    def crop_url(self):
        env = self.env
        # XXX looks like a hack
        if any(x for x in env.stream.actions if x.action=="image_upload"):
            return env.stream.url_for(env, 'image_upload.crop',
                                      item=self.form.item.id,
                                      field_name=self.input_name)

    @cached_property
    def name_parent(self):
        parent = self.parent
        while parent is not None:
            if isinstance(parent, Form) or getattr(parent, 'name', None):
                return parent
            parent = getattr(parent, 'parent', None)
        return None

    @cached_property
    def model(self):
        parent = self.name_parent
        if isinstance(parent, Form):
            if parent.item:
                return parent.item.__class__
            return parent.model
        elif isinstance(parent.conv, convs.ModelDictConv):
            if parent.clean_value is not None:
                return parent.clean_value.__class__
            return parent.conv.model
        else:
            return None

    @cached_property
    def model_field(self):
        if self.model is None:
            return None

        model_field = getattr(self.model, self.name)
        if isinstance(model_field, FileAttribute) and \
           isinstance(model_field.prop, ImageProperty):
            return model_field.prop

    @cached_property
    def size(self):
        if self.model_field is not None:
            return self.model_field.image_sizes

    @cached_property
    def thumb_size(self):
        if self.retina and self.size is not None:
            w, h = self.size
            return w/2, h/2
        return self.size

    @cached_property
    def label(self):
        if self.size is not None and self._label and self.show_size:
            return self._label + u' ({}×{})'.format(*self.size)
        return self._label

    @cached_property
    def fill_from(self):
        if self.model_field is not None:
            fill_from = self.model_field.fill_from
            if fill_from is not None:
                if '.' in self.input_name:
                    parent_name = self.input_name.rsplit('.', 1)[0]
                    fill_from = parent_name + '.' + fill_from

                if self.form.get_field(fill_from):
                    return fill_from

        return None

    def set_raw_value(self, raw_data, value):
        AjaxFileField.set_raw_value(self, raw_data, value)
        # XXX hack!
        if getattr(self.form, 'for_diff', False) and \
                self.clean_value and \
                not self.conv.autocrop:
            resizer = ResizeFit()
            try:
                img = Image.open(self.clean_value.path)
            except IOError:
                pass
            else:
                img = resizer(img, (100, 100))
                img = img.convert('RGB')
                img_file = StringIO()
                img.save(img_file, format='jpeg')
                data = "data:image/jpeg;base64," + \
                        img_file.getvalue().encode('base64').replace('\n', '')
                raw_data[self.prefix+'image'] = data

    def get_diff(field1, field2):
        path1 = field1.form.raw_data.get(field1.prefix+'path')
        path2 = field2.form.raw_data.get(field2.prefix+'path')
        if path1 != path2:
            img1 = field1.form.raw_data.get(field1.prefix+'image')
            img2 = field2.form.raw_data.get(field2.prefix+'image')
            if img1 or img2:
                if img1:
                    before = lambda: Markup("<img src='{}'/>").format(img1)
                else:
                    before = lambda: ''
                if img2:
                    after = lambda: Markup("<img src='{}'/>").format(img2)
                else:
                    after = lambda: ''
                return dict(label=field2.label,
                            name=field2.input_name,
                            before=before,
                            after=after,
                            changed=True)
            elif not field1.conv.autocrop and not field2.conv.autocrop:
                return dict(label=field2.label,
                            name=field2.input_name,
                            before=lambda: path1,
                            after=lambda: path2,
                            changed=True)






# -*- coding: utf-8 -*-
'''Instrument for locking individual objects by one and only one editor in time'''

from iktomi.cms.item_lock.memcached import MemcachedItemLock as ItemLock # default
from iktomi.cms.item_lock.base import (ItemLockData, ModelLockError, \
        ModelLockedByOther, ModelLockIsLost, logger)



# -*- coding: utf-8 -*-
from iktomi.db.sqla.files import FileProperty, FileEventHandlers
from iktomi.db.sqla.images import ImageProperty, ImageEventHandlers
from sqlalchemy.orm import object_session

from .model import _FrontReplicated, _WithState

# ======================================================================

class ReplicatedHandlersMixin(object):

    def _get_file_name_to_delete(self, target, changes):
        if changes and changes.deleted:
            old_name = changes.deleted[0]
            if old_name != getattr(target._other_version, self.prop.column.key):
                return old_name


class PublicatedHandlersMixin(object):

    def before_update(self, mapper, connection, target):
        super(PublicatedHandlersMixin, self)\
                                    .before_update(mapper, connection, target)
        self.update_symlinks(target)

    def update_symlinks(self, target):
        if isinstance(target, _FrontReplicated):
            if isinstance(target, _WithState):
                session = object_session(target)
                source = target._admin_item
                attr = getattr(type(target), self.prop.key)
                target_file = getattr(target, self.prop.key)
                source_file = getattr(source, self.prop.key)
                if target_file is not None:
                    file_manager = session.find_file_manager(target)
                    if target.public:
                        file_manager.create_symlink(source_file, target_file)
                    else:
                        file_manager.delete(target_file)


class ReplicatedImageEventHandlers(ReplicatedHandlersMixin, ImageEventHandlers):
    pass

class ReplicatedFileEventHandlers(ReplicatedHandlersMixin, FileEventHandlers):
    pass

class PublicatedFileEventHandlers(
    PublicatedHandlersMixin,
    ReplicatedHandlersMixin,
    FileEventHandlers
): pass

class PublicatedImageEventHandlers(
    PublicatedHandlersMixin,
    ReplicatedHandlersMixin,
    ImageEventHandlers
): pass



# XXX name of class in too long
class ReplicatedFileProperty(FileProperty):
    event_cls = ReplicatedFileEventHandlers


class ReplicatedImageProperty(ImageProperty):
    event_cls = ReplicatedImageEventHandlers


class PublicatedFileProperty(FileProperty):
    event_cls = PublicatedFileEventHandlers


class PublicatedImageProperty(ImageProperty):
    event_cls = PublicatedImageEventHandlers




# -*- coding: utf-8 -*-
from webob.exc import HTTPNotFound
from iktomi import web
from iktomi.cms.stream_handlers import PrepareItemHandler
#from iktomi.unstable.db.sqla.replication import replicate_attributes
from iktomi.cms.publishing.stream import PublishItemHandler, \
        PublishStreamNoState, PublishStream
from iktomi.cms.stream import Stream
from iktomi.cms.forms import convs
from iktomi.utils import cached_property


class PrepareI18nItemHandler(PrepareItemHandler):

    def prepare_item_handler(self, env, data):
        # XXX Dirty hack to support object creation
        env.absent_items = True
        return PrepareItemHandler.__call__(self, env, data)
    __call__ = prepare_item_handler


class I18nItemHandler(PublishItemHandler):

    # XXX turn off autosave or make it save to DraftForm only?

    PrepareItemHandler = PrepareI18nItemHandler

    def drop_field_on_i18n(self, field, key):
        return field and (isinstance(field.conv, convs.Char) or
                          field.name == 'id')

    def get_item_form(self, stream, env, item, initial, draft=None):
        if item.state not in (item.ABSENT, item.DELETED):
            return PublishItemHandler.get_item_form(
                    self, stream, env, item, initial, draft)

        # XXX this method looks hacky
        # Get existing language version and fill the form with object reflection
        # to current language model
        for lang in item.models.langs:
            # XXX item.models is not an interface
            if lang == item.models.lang:
                continue
            source_item = item._item_version('admin', lang)
            if source_item and \
                    source_item.state not in (item.ABSENT, item.DELETED):
                break
        else:
            # The item has been deleted on all language versions, creation is
            # not allowed
            raise HTTPNotFound
        # make object reflection, do not add it to db
        #fake_item = item.__class__()
        ## XXX do not replicate text fields, creation time, etc
        #replicate_attributes(source_item, fake_item)
        #form = PublishItemHandler.get_item_form(
        #        self, stream, env, fake_item, initial, draft)
        # XXX hack!
        #form.item = item

        # hack do get initial value for form from source item
        source_form = PublishItemHandler.get_item_form(
                    self, stream, env, source_item, initial, draft)
        form = PublishItemHandler.get_item_form(
                self, stream, env, item, initial, draft)
        form.accept(source_form.raw_data)
        form.errors = {}
        for key in form.raw_data.keys():
            field = form.get_field(key)
            if self.drop_field_on_i18n(field, key):
                field.set_raw_value(form.raw_data,
                                    field.from_python(field.get_initial()))
        return form

    def process_item_template_data(self, env, td):
        item = td['item']
        if item.state in (item.ABSENT, item.DELETED):
            td['title'] = u'Создание языковой версии объекта'
        return PublishItemHandler.process_item_template_data(self, env, td)


class I18nStreamMixin(object):

    langs = [('ru', u'Русский'),
             ('en', u'Английский')]

    @cached_property
    def langs_dict(self):
        return dict(self.langs)

    # If any item in one language have a pair in other language
    # Basically, if it is instance of iktomi.cms.publishing.model.WithLanguage
    items_are_i18n = True

    list_base_template = 'lang_publish_stream.html'

    def uid(self, env, version=True):
        # Attention! Be careful!
        # Do not change format of uid unless you are sure it will not 
        # brake tray views, where stream_name and language are parsed out
        # from the uid
        if version:
            return self.module_name + ':version=' + env.version + ':lang=' + env.lang
        return self.module_name + ':lang=' + env.lang

    @property
    def prefix_handler(self):
        @web.request_filter
        def set_models(env, data, nxt):
            #assert data.version in self.versions_dict.keys()
            env.models = getattr(env.models, data.version)
            env.models = getattr(env.models, data.lang)
            env.version = data.version
            env.lang = data.lang
            return nxt(env, data)

        version_prefix = web.prefix('/<any("%s"):version>' % \
                                     ('","'.join(self.versions_dict.keys())))
        lang_prefix = web.prefix('/<any("%s"):lang>' % \
                                     ('","'.join(self.langs_dict.keys())))
        #return version_prefix | set_models | \
        return super(PublishStreamNoState, self).prefix_handler |\
               version_prefix | lang_prefix | set_models

    def url_for(self, env, name=None, **kwargs):
        kwargs.setdefault('version', getattr(env, 'version', None) or self.versions[0][0])
        kwargs.setdefault('lang', getattr(env, 'lang', None) or self.langs[0][0])
        return Stream.url_for(self, env, name, **kwargs)


class I18nPublishStreamNoState(I18nStreamMixin, PublishStreamNoState):
    pass


class I18nPublishStream(I18nStreamMixin, PublishStream):

    core_actions = [x for x in PublishStream.core_actions 
                    if x.action != 'item'] + [
        I18nItemHandler(),
    ]




# -*- coding: utf-8 -*-

import logging, warnings

from webob.exc import HTTPForbidden
from webob.multidict import MultiDict
from jinja2 import Markup

from iktomi.utils import cached_property
from collections import OrderedDict
from iktomi.utils.storage import VersionedStorage
from iktomi import web
from iktomi.web.url_converters import Converter, default_converters
from iktomi.cms.forms import Form
from iktomi.cms.item_lock import ItemLock
from iktomi.utils.deprecation import deprecated
from . import stream_handlers as handlers
from .flashmessages import flash

logger = logging.getLogger(__name__)


_none_converter_factory_cache = {}

def none_converter_factory(cls):
    if cls in _none_converter_factory_cache:
        return _none_converter_factory_cache[cls]

    class NoneConverter(cls):
        regex = Converter.regex

        def to_python(self, value, **kwargs):
            if value == '+':
                return None
            return super(NoneConverter, self).to_python(value, **kwargs)

        def to_url(self, value):
            if value is None:
                return '+'
            return super(NoneConverter, self).to_url(value)

    _none_converter_factory_cache[cls] = NoneConverter

    return NoneConverter


def I18nLabel(string, lang):
    return Markup(u'<span class="lang-%s">%s</span>' % (lang, string))


def ListFields(*args):
    fields = []
    for field in args:
        if not isinstance(field, ListField):
            field = ListField(*field)
        fields.append(field)
    if not fields or fields[0].name != 'id':
        fields.insert(0, ListField('id', 'ID', classname='width-0'))
    fields.insert(1, ItemLockListField())
    return OrderedDict(fields)


class ListField(object):

    template = 'list_field.html'

    # Use classname and define width in css files
    _obsolete = frozenset(['width'])

    def __init__(self, name, title, image=False,
                 transform=lambda f: u'—' if f is None else f,
                 static=False, link_to_item=True, classname='',
                 template=None, **kwargs):
        if self._obsolete & set(kwargs):
            raise TypeError('Obsolete parameters are used: {}'.format(
                list(self._obsolete & set(kwargs))))
        self.name = name
        self.title = title
        self.image = image
        self.static = static
        self.transform = transform
        self.link_to_item = link_to_item
        self.classname = classname
        self.template = template or self.template

    def __iter__(self):
        yield self.name
        yield self

    def __call__(self, env, item, url, loop):
        field_val = self.get_value(env, item, url, loop)
        if self.transform is not None:
            return self.transform(field_val)
        return field_val

    def get_value(self, env, item, url, loop):
        return getattr(item, self.name)


class ItemLockListField(ListField):

    template = 'list_field_item_lock.html'

    def __init__(self, name='locked', title='', **kwargs):
        kwargs.setdefault('link_to_item', False)
        kwargs.setdefault('transform', lambda x: x)
        kwargs.setdefault('classname', 'width-0')
        ListField.__init__(self, name, title, **kwargs)

    def get_value(self, env, item, url, loop):
        lock = env.item_lock.check(item)
        if lock is not None:
            return dict(lock,
                        guid=env.item_lock.item_global_id(item),
                        user=env.db.query(env.auth_model)\
                                   .get(lock['user_id']))


class FilterForm(Form):

    fields = []

    js_block = 'filter-form'
    search_input = None # for live search widget

    def filter_by_scalar(self, query, field, value):
        return query.filter(getattr(self.model, field.name)==value)

    def filter_by_true(self, query, field, value):
        if value:
            return self.filter_by_scalar(query, field, value)
        return query

    def filter_by_list(self, query, field, values):
        prop = getattr(self.model, field.name)
        for value in values:
            query = query.filter(prop.contains(value))
        return query

    def filter_by_default(self, query, field, value):
        if field.multiple:
            return self.filter_by_list(query, field, value)
        else:
            return self.filter_by_scalar(query, field, value)

    def filter(self, query):
        '''Modifies query'''
        # XXX will not work with FieldBlocks!
        for field in self.fields:
            filter_value = self.python_data[field.name]
            if filter_value or filter_value == 0:
                method = getattr(self, 'filter_by__%s' % field.name,
                                 getattr(field, 'filter_query',
                                         self.filter_by_default))
                query = method(query, field, filter_value)
        return query

    def defaults(self):
        return {}

    def get_data(self, compact=True):
        data = MultiDict(self.raw_data) # XXX
        if compact:
            compact_data = MultiDict()
            for key, value in data.iteritems():
                if value:
                    compact_data.add(key, value)
            data = compact_data
        return data

    def __nonzero__(self):
        # We don't want to display form when there is no fields
        return bool(self.fields)


class Stream(object):
    id_converter_name = 'int'

    actions = []
    core_actions = [handlers.StreamListHandler(),
                    handlers.EditItemHandler(),
                    handlers.DeleteItemHandler(),
                    handlers.GetReferrersHandler(),
                    ]

    buttons = ['save', 'save_and_continue', 'save_and_add_another', 'delete']

    lock_back_title = None
    lock_back_help = None

    def __init__(self, module_name, config):
        self.config = config
        self.module_name = module_name
        self.actions = [x.bind(self) for x in self.core_actions + self.actions]
        self.core_actions = []

    @cached_property
    def id_converter(self):
        return default_converters.get(self.id_converter_name)

    @cached_property
    def id_none_converter(self):
        return none_converter_factory(self.id_converter)

    @property
    def prefix_handler(self):
        """ Handler match stream path, setup stream namespace and env"""
        @web.request_filter
        def set_stream_handler(env, data, nxt):
            env.stream = self
            return nxt(env, data)

        part = self.module_name.rsplit('.', 1)[-1]
        return web.prefix('/' +part, name=part) | \
               set_stream_handler

    @property
    def app_handler(self):
        """ Handler add stream action's app handlers """
        apps = [action.app for action in self.actions]
        return web.cases(*apps)

    def get_handler(self):
        """ Get web handler for routing.
        """
        return self.prefix_handler | self.app_handler

    def url_for(self, env, name=None, **kwargs):
        name = name and '%s.%s' % (self.module_name, name) or self.module_name
        return env.url_for(name, **kwargs)

    def get_edit_url(self, env, item):
        '''
        Checks if item belongs to the stream, and if it's true,
        returns an url to item edit page
        '''
        # be careful, this method is redefined for publish streams
        if isinstance(item, self.get_model(env)):
            #cls, id = identity_key(instance=item)
            item_ = self.item_query(env).filter_by(id=item.id).scalar()
            if item_ is not None:
                return self.url_for(env, 'item', item=item.id)

    @cached_property
    def autosave(self):
        return getattr(self.config, 'autosave', True)

    @cached_property
    @deprecated("use `stream.edit_log_action is not None` instead")
    def edit_log(self):
        return any(x for x in self.actions if x.action=='edit_log')

    @cached_property
    def referrers(self):
        return any(x for x in self.actions if x.action=='referrers')

    @cached_property
    def app_namespace(self):
        if '.' in self.module_name:
            return self.module_name.rsplit('.', 1)[0]
        return ''

    @cached_property
    def perms(self):
        p = getattr(self.config, 'permissions', {})
        p.setdefault('wheel', 'rwxcdp')
        return p

    @cached_property
    def edit_action(self):
        for action in self.actions:
            if action.action=='item':
                return action

    @cached_property
    def list_edit_action(self):
        for action in self.actions:
            if hasattr(action, 'ListItemForm'):
                return action

    @cached_property
    def preview_action(self):
        for action in self.actions:
            if action.action=='preview':
                return action

    @cached_property
    def edit_log_action(self):
        for action in self.actions:
            if action.action=='edit_log':
                return action

    def create_log_entry(self, env, item, type_):
        EditLog = getattr(env, 'edit_log_model', None)
        log_enabled = EditLog is not None and \
                      self.edit_log_action is not None
        if log_enabled:
            return EditLog(stream_name=self.uid(env),
                           type=type_,
                           object_id=item.id,
                           global_id=ItemLock.item_global_id(item),
                           users=[env.user])

    @cached_property
    def ListItemForm(self):
        if self.list_edit_action:
            return self.list_edit_action.ListItemForm

    @cached_property
    def stream_template_name(self):
        if hasattr(self.config, 'stream_template'):
            return self.config.stream_template
        if hasattr(self.config, 'template'):
            warnings.warn("Stream.config.template is deprecated",
                          category=DeprecationWarning)
            return self.config.template
        return 'stream'

    @cached_property
    def row_template_name(self):
        return getattr(self.config, 'row_template', 'stream_row')

    @cached_property
    def item_template_name(self):
        return getattr(self.config, 'item_template', 'item')

    @cached_property
    def template_data(self):
        return getattr(self.config, 'template_data', {})

    @cached_property
    def live_search(self):
        return getattr(self.config, 'live_search', False)

    @cached_property
    def list_fields(self):
        return getattr(self.config, 'list_fields', {})

    @cached_property
    def title(self):
        return getattr(self.config, 'title', self.module_name)

    def uid(self, env, version=None):
        return self.module_name

    def get_filter_form(self, env):
        cls = getattr(self.config, 'FilterForm', FilterForm)
        form = cls(env)
        form.model = self.get_model(env)
        return form

    def process_item_template_data(self, env, template_data):
        '''Preprocessor for template variables.
           Can be overriden by descedant classes.'''
        return template_data

    def process_list_template_data(self, env, template_data):
        '''Preprocessor for template variables.
           Can be overriden by descedant classes.'''
        return template_data

    def lock_back_url(self, env, item, filter_form):
        return self.url_for(env).qs_set(filter_form.get_data())

    def order(self, query):
        return query

    def get_permissions(self, env):
        perms = set()
        for role in env.user.roles:
            # XXX how does it work for wheel?
            perms |= set(self.perms.get(role, ''))
        return perms

    def has_permission(self, env, permission):
        return permission in self.get_permissions(env)

    def insure_has_permission(self, env, permission):
        if not self.has_permission(env, permission):
            raise HTTPForbidden

    #def stream_endpoint(self, env):
    #    return self.module_name
    #def stream_url(self, request):
    #    return request.url_for(self.stream_endpoint(request))

    def get_model(self, env):
        return self.config.Model

    def __repr__(self):
        return '<%s.%s: %s>' % (self.__class__.__module__, self.__class__.__name__, self.module_name)

    def item_query(self, env):
        return env.db.query(self.get_model(env))

    # ========= Item actions ====

    def commit_item_transaction(self, env, item, silent=False):
        '''commits request.db and flashes success message'''
        env.db.commit()
        if not silent:
            flash(env, u'Объект (%s) сохранен' % (item,), 'success')

    def rollback_due_lock_lost(self, env, item, silent=False):
        '''rollbacks request.db and flashes failure message'''
        env.db.rollback()
        if not silent:
            flash(env, u'Объект (%s) не был сохранен из-за '
                       u'перехваченной блокировки' % (item,),
                       'failure')

    def rollback_due_form_errors(self, env, item, silent=False):
        env.db.rollback()
        if not silent:
            flash(env, u'Объект (%s) не был сохранен из-за ошибок' % (item,),
                       'failure')


def decode_stream_uid(stream_name):
    if ':' in stream_name:
        name = stream_name.split(':', 1)[0]
        params = dict([x.split('=', 1) for x in
                       stream_name.split(':')[1:]
                       if '=' in x])
        return name, params
    return stream_name, {}


def expand_stream(env, obj):
    stream_name, params = decode_stream_uid(obj.stream_name)

    stream_env = VersionedStorage()
    stream_env._storage._parent_storage = env

    if 'lang' in params:
        stream_env.models = getattr(env.models, params['lang'])
        stream_env.lang = params['lang']
    if stream_name not in env.streams:
        # Deleted or renamed stream
        raise NotImplementedError
    stream = env.streams[stream_name]
    item = stream.item_query(stream_env)\
                 .filter_by(id=obj.object_id).first()
    if item is not None:
        stream_title = stream.title
        if 'lang' in params:
            stream_title = I18nLabel(stream_title, params['lang'])
        url = stream.url_for(stream_env, 'item', 
                             item=obj.object_id, **params)
        return (url, stream_title, stream, obj, item)





# -*- coding: utf-8 -*-

from sqlalchemy import Column, ForeignKey, Integer, DateTime, \
        Index, String, Text
from sqlalchemy.orm import relationship
from datetime import datetime
from iktomi.cms.models.base import register_model

__all__ = ['EditorNote']

@register_model('BaseModel')
def EditorNote(models):

    id = Column(Integer, primary_key=True)
    stream_name = Column(String(50), nullable=False, default='')
    # object id can be string, so we use string here
    object_id = Column(String(50), nullable=True)
    created_dt = Column(DateTime, default=datetime.now, nullable=False)
    editor_id = Column(Integer, ForeignKey(models.AdminUser.id), nullable=False)
    editor = relationship(models.AdminUser)
    body = Column(Text, nullable=False)

    @classmethod
    def get_for_item(cls, db, stream_name, item):
        query = db.query(cls).filter_by(stream_name=stream_name)
        return query.filter(cls.object_id == str(item.id)).all()

    Index('draft_index', stream_name, object_id)



# -*- coding: utf-8 -*-
from iktomi.utils import cached_property, weakproxy
from iktomi.utils.storage import VersionedStorage
from iktomi import web
#from iktomi.cms.stream_handlers import insure_is_xhr


class Menu(object):

    template = 'menu/menu'
    rel = None

    def __init__(self, title, link=None, endpoint=None, params=None,
                 items=None, env=None, template_vars={}, template=None,
                 permissions= {'*': 'rwxcd'}, rel=None):
        self.parent = None
        self.title = title
        self.link = link
        self.endpoint = endpoint
        self.params = params or {}
        self.items = filter(None, items or [])
        self._env = env
        self.template = template or self.template
        self.template_vars = template_vars
        self.perms = {'wheel': 'rwxcd'}
        self.perms.update(permissions)
        for item in self.items:
            item.parent = weakproxy(self)
        self.rel = self.rel or rel

    @cached_property
    def env(self):
        return self._env or (self.parent.env if self.parent else None)

    @cached_property
    def request(self):
        return self.env.request

    @cached_property
    def get_permissions(self):
        if self.url is None and not self.has_children:
            return ''
        perms = set(self.perms.get('*', ''))
        if self.env is None:
            return perms
        for role in self.env.user.roles:
            perms |= set(self.perms.get(role, ''))
        return perms

    @cached_property
    def has_children(self):
        for item in self.items:
            if len(item.get_permissions) > 0:
                return True
        return False

    @cached_property
    def url(self):
        if self.link:
            return self.link
        if self.endpoint:
            return self.env.url_for(self.endpoint, **self.params)
        return None

    @cached_property
    def active(self):
        return self.request.path == self.url

    @cached_property
    def endpoint_name(self):
        return self.endpoint

    def __iter__(self):
        for item in self.items:
            yield item

    def render(self):
        if len(self.get_permissions) == 0:
            return ''
        # check if allowed menu has no allowed children nor direct link
        if self.url is None and not self.has_children:
            return ''
        return self.env.render_to_string(self.template, dict(
            menu=self, url=self.url, active=self.active,
            title=self.title, rel=self.rel, **self.template_vars))

    def child_count(self):
        return len(self.items)

    def max_childs(self):
        return reduce(max, [x.child_count() for x in self.items], 1)


class StreamMenu(Menu):

    template = 'menu/stream_menu'

    def __init__(self, stream_name, title=None, create=True,
                 filters=None, items=None, template_vars={},
                 env=None, template=None):
        self.stream_name = stream_name
        if title is not None:
            self.title = title
        self.create = create
        self.filters = filters or {}
        self.items = items or []
        self.template = template or self.template
        self.template_vars = template_vars
        self._env = env
        for item in self.items:
            item.parent = weakproxy(self)

    @cached_property
    def get_permissions(self):
        return self.stream.get_permissions(self.env)

    @cached_property
    def active(self):
        if self.filters:
            return self.url == '%s?%s' % (self.request.path,
                                          self.request.query_string)
        return self.url == self.request.path

    @cached_property
    def stream(self):
        return self.env.streams[self.stream_name]

    @cached_property
    def title(self):
        return self.stream.title

    @cached_property
    def url(self):
        return self.stream.url_for(self.env).qs_set(self.filters)

    @cached_property
    def create_url(self):
        return self.stream.url_for(self.env, 'item', item=None)\
                                                .qs_set(self.filters)

    @cached_property
    def endpoint_name(self):
        return self.stream_name


class ActionMenu(Menu):

    def __init__(self, stream_name, action, title=None, filters=None):
        Menu.__init__(self, title)
        self.stream_name = stream_name
        self.action = action
        self.filters = filters or {}

    @cached_property
    def active(self):
        if self.filters:
            return self.url == '%s?%s' % (self.request.path,
                                          self.request.query_string)
        return self.url == self.request.path

    @cached_property
    def stream(self):
        return self.env.streams[self.stream_name]

    @cached_property
    def url(self):
        return self.stream.url_for(self.env, self.action)(self.filters)


class LonerMenu(StreamMenu):

    #template = 'menu/loner_menu'

    def __init__(self, stream_name, title=None, items=None,
                 template_vars={}, env=None, template=None):
        StreamMenu.__init__(self, stream_name,
                            title=title, items=items,
                            template_vars=template_vars, env=env,
                            template=template)
        self.filters = {}
        self.create = False


class DashCol(Menu):

    template = 'menu/dashboard-col'


class MenuGroup(Menu):

    def __init__(self, items, env=None, template=None):
        Menu.__init__(self, None, items=items, env=env, template=template)


class LangStreamMenu(StreamMenu):

    def __init__(self, *args, **kwargs):
        self.lang = kwargs.pop('lang')
        StreamMenu.__init__(self, *args, **kwargs)

    @cached_property
    def env(self):
        # XXX hack
        env = super(LangStreamMenu, self).env
        vs = VersionedStorage(lang=self.lang)
        vs._storage._parent_storage = env
        return vs


class LangLonerMenu(LonerMenu):
    # XXX DRY

    def __init__(self, *args, **kwargs):
        self.lang = kwargs.pop('lang')
        LonerMenu.__init__(self, *args, **kwargs)

    @cached_property
    def env(self):
        # XXX hack
        env = super(LangLonerMenu, self).env
        vs = VersionedStorage(lang=self.lang)
        vs._storage._parent_storage = env
        return vs


class DashRow(MenuGroup):

    template = 'menu/dashboard'


def DashI18nStream(*args, **kwargs):
    return MenuGroup([LangStreamMenu(*args, **dict(kwargs, lang=lang))
                      for lang in ('ru', 'en')],
                      template="menu/dashboard-row-i18n")

def DashI18nLoner(*args, **kwargs):
    return MenuGroup([LangLonerMenu(*args, **dict(kwargs, lang=lang))
                      for lang in ('ru', 'en')],
                      template="menu/dashboard-row-i18n")

def DashStream(*args, **kwargs):
    kwargs.setdefault('template', 'menu/dashboard-row')
    return StreamMenu(*args, **kwargs)

def DashMenu(*args, **kwargs):
    kwargs.setdefault('template', 'menu/dashboard-row')
    return Menu(*args, **kwargs)

def DashLoner(*args, **kwargs):
    kwargs.setdefault('template', 'menu/dashboard-row')
    return LonerMenu(*args, **kwargs)


class IndexHandler(web.WebHandler):
    "A view for index page"

    def __init__(self, dashboard):
        self.dashboard = dashboard

    def index(self, env, data):
        #insure_is_xhr(env)

        return env.render_to_response('index', dict(
            title=u'Редакторский интерфейс сайта',
            menu='index',
            dashboard=self.dashboard(env),
        ))
    __call__ = index







# -*- coding: utf-8 -*-
from importlib import import_module
from iktomi.cms.stream import Stream
from iktomi import web


class Streams(dict):

    @classmethod
    def from_list(cls, stream_list, package, **kwargs):
        return cls((name, cls.get_stream(name, package, **kwargs)) \
                   for name in stream_list)

    @classmethod
    def from_tree(cls, stream_tree, package, **kwargs):
        stream_list = []
        for stream in stream_tree:
            if type(stream) in (list, tuple):
                stream_list += [stream[0] + '.' + s for s in stream[1:]]
            else:
                stream_list.append(stream)
        return cls.from_list(stream_list, package, **kwargs)

    @staticmethod
    def get_stream(name, package, stream_class=Stream):
        module = import_module('.' + name, package)
        stream_class = getattr(module, 'Loner', stream_class)
        stream_class = getattr(module, 'Stream', stream_class)
        return stream_class(name, module)

    def to_app(self):
        return self._create_subapp()

    def _create_subapp(self, ns=''):
        prefix = ns and (ns + '.') or ''
        handlers_by_namespace = self._handlers_by_namespace()

        children = filter(
            lambda x: x and x.startswith(prefix) and not '.' in x[len(prefix):],
            handlers_by_namespace)

        child_apps = [web.prefix('/' + x.rsplit('.', 1)[-1]) | \
                      web.namespace(x.rsplit('.', 1)[-1]) | 
                      self._create_subapp(x)
                      for x in children]
        return web.cases(*(child_apps + handlers_by_namespace[ns]))

    def _handlers_by_namespace(self):
        result = {}
        for stream in self.values():
            for namespace in self._get_parent_namespaces(stream.app_namespace):
                result.setdefault(namespace, [])
            result[stream.app_namespace].append(stream.get_handler())
        return result

    def _get_parent_namespaces(self, namespace):
        result = []
        parts = namespace.split('.')
        while parts:
            result.append('.'.join(parts))
            parts.pop()
        return result





# -*- coding: utf-8 -*-
# Common admin views
from webob.exc import HTTPNotFound, HTTPBadRequest, HTTPForbidden
from iktomi import web
from iktomi.cms.stream import expand_stream
from iktomi.web.shortcuts import Rule
from iktomi.cms.forms import Form, convs
from iktomi.cms.forms.fields import Field, object_ref_fields


class TrayView(web.WebHandler):

    def __init__(self, Tray, ObjectTray, AdminUser):
        self.Tray = Tray
        self.ObjectTray = ObjectTray

        class MyTrayObjectForm(Form):

            fields = object_ref_fields + [
                Field('comment',
                      conv=convs.Char(required=False)),
            ]

        class TrayObjectForm(Form):

            fields = MyTrayObjectForm.fields + [
                Field('tray',
                      conv=convs.ModelChoice(model=Tray, required=True)),
            ]

        class UserTrayObjectForm(Form):

            fields = MyTrayObjectForm.fields + [
                Field('user',
                      conv=convs.ModelChoice(model=AdminUser,
                                             condition=AdminUser.active==True,
                                             required=True)),
            ]

        self.TrayObjectForm = TrayObjectForm
        self.MyTrayObjectForm = MyTrayObjectForm
        self.UserTrayObjectForm = UserTrayObjectForm

    @property
    def app(self):
        return web.prefix('/tray') | web.cases(
            Rule('/<int:tray>', self.tray),
            web.prefix('/_') | web.method('POST', strict=True) | web.cases(
                Rule('/put', self.put_to_tray),
                Rule('/user/put', self.put_to_user_tray),
                Rule('/delete', self.delete_from_tray),
            )
        )

    def put_to_tray(self, env, data):
        form = self.TrayObjectForm(env)
        if not form.accept(env.request.POST):
            return env.json({'success': False,
                             'errors': form.errors})
        return self._put_to_tray(env, **form.python_data)

    def put_to_user_tray(self, env, data):
        form = self.UserTrayObjectForm(env)
        if not form.accept(env.request.POST):
            return env.json({'success': False,
                             'errors': form.errors})
        return self._put_to_user_tray(env, form.python_data['user'], form)

    def _put_to_user_tray(self, env, user, form):
        tray = env.db.query(self.Tray).filter_by(editor=user).first()
        if tray is None:
            tray = self.Tray(editor=user,
                             title=user.name or user.login)
            env.db.add(tray)
            env.db.flush()
        return self._put_to_tray(env, tray=tray, **form.python_data)

    def _put_to_tray(self, env, stream_name=None, object_id=None,
                     tray=None, comment=None, user=None):
        filter_args = dict(stream_name=stream_name,
                           object_id=object_id,
                           tray=tray)
        object_tray = env.db.query(self.ObjectTray)\
                            .filter_by(**filter_args).first()
        if object_tray is not None:
            return env.json({'success': False,
                             'error': u'Объект уже находится в папке'})
        object_tray = self.ObjectTray(**filter_args)
        env.db.add(object_tray)
        object_tray.comment = comment
        object_tray.sender = env.user
        env.db.commit()
        return env.json({'success': True,
                         'id': object_tray.id,
                         'tray_id': tray.id,
                         'title': tray.title})

    def delete_from_tray(self, env, data):
        try:
            id = int(env.request.POST.get('id', ''))
        except ValueError:
            raise HTTPBadRequest()
        obj = env.db.query(self.ObjectTray).get(id)
        if obj is None:
            raise HTTPBadRequest()
        if not obj.can_delete(env.user):
            raise HTTPForbidden()
        env.db.delete(obj)
        env.db.commit()
        return env.json({'success': True})

    # XXX deleting on GET 
    def clear_tray(self, env, obj_items):
        for obj, item in obj_items:
            if item is None:
                env.db.delete(obj)
        env.db.commit()

    def tray(self, env, data):
        #insure_is_xhr(env)
        env.models = env.models.admin
        env.version = 'admin'
        tray = env.db.query(self.Tray).get(data.tray)
        if tray is None:
            raise HTTPNotFound()
        objects = env.db.query(self.ObjectTray).filter_by(tray=tray).all()
        items = [expand_stream(env, obj) for obj in objects]
        if None in items:
            self.clear_tray(env, zip(objects, items))
        items = [item for item in items if item is not None]
        #changed.sort(key=lambda x: x.date_changed)
        return env.render_to_response('tray', dict(
            tray = tray,
            items = items,
            menu = env.current_location,
            title = tray.title,
        ))



# -*- coding: utf-8 -*-
from iktomi.forms.convs import *
from iktomi.forms.convs import __all__ as _all1
#_all1 = locals().keys()
from sqlalchemy.orm import Query
from iktomi.cms.publishing.model import WithState
from datetime import date
_all2 = locals().keys()


def DBUnique(model=None, message=u'An object with such value already exists'):
    '''
      Checks if an object with the same value does not exist in db,
      except the current object.

      Be careful with unique indexes in published models. They are not actually
      deleted from database on delete action, but their state is changed
      to DELETED. So they may exist in index, but be ignored by DBUnique, 
      and you will get an exception.
    '''

    @validator(message)
    def to_python(conv, value):
        model_ = model or conv.field.form.model
        if value not in ('', None):
            field = getattr(model_, conv.field.name)
            # XXX Using db.query() won't work properly with custom
            # auto-filtering query classes. But silently replacing the class is
            # not good too.
            query = Query(model_, session=conv.env.db)
            if issubclass(model_, WithState):
                states = WithState.PRIVATE, WithState.PUBLIC
                query = query.filter(model_.state.in_(states))
            item = query.filter(field==value).scalar()
            if item is not None and item != conv.field.form.item:
                return False
        return True
    return to_python


class DateIntervalConv(Converter):

    def to_python(self, value):
        if type(value['since']) is date \
                and type(value['till']) is date \
                and value['since'] > value['till']:
            raise ValidationError(
                        u'дата начала не может быть больше даты завершения')
        return value

# Expose all variables defined after imports and all variables imported from
# parent module
__all__ = [x for x
           in set(locals().keys()) - (set(_all2) - set(_all1))
           if not x.startswith('_')]
del _all1, _all2



# -*- coding: utf-8 -*-
import time
import json
from webob import Response
from iktomi import web


def set_flash_cookies(env, result):
    if getattr(env, '_flash', None) and isinstance(result, Response):
        result.set_cookie('flash-msg-%s' % time.time(),
                          json.dumps(env._flash), max_age=120)
    return result


@web.request_filter
def flash_message_handler(env, data, nxt):
    env._flash = []
    result = nxt(env, data)
    return set_flash_cookies(env, result)


def flash(env, message, category=None, unique=True):
    # XXX this will not work on errors
    if not hasattr(env, '_flash'):
        env._flash = []
    value = (unicode(message), category)
    if not unique or value not in env._flash:
        env._flash.append(value)



# -*- coding: utf-8 -*-
from iktomi.cms.stream import FilterForm
from iktomi.cms.forms import convs, widgets
from iktomi.cms.forms.fields import Field, DateFromTo, SortField


#def all_by_default(conv, value):
#    if not value:
#        return [k for k, v in conv.conv.choices]
#    return value


class EditLogItemFilterForm(FilterForm):

    def __init__(self, *args, **kwargs):
        # XXX hack! do not do in this way. Use class_factory
        streams = kwargs.pop('streams', None)
        types = kwargs.pop('types')
        models = kwargs.pop('models')
        env = args[0]
        fields = []
        for field in self.fields:
            field = field()
            if field.name == 'users':
                AdminUser = env.auth_model
                field = field(conv=convs.ModelChoice(model=AdminUser,
                                                     title_field="login"))
            if field.name == 'streams':
                field = field(conv=convs.ListOf(
                            convs.EnumChoice(choices=streams),
                            #all_by_default
                            ))
            if field.name == 'type':
                field = field(conv=convs.EnumChoice(choices=types))
            fields.append(field)
        self.fields = fields
        FilterForm.__init__(self, *args, **kwargs)

    fields = [
        Field('object_id',
              label=u"ID",
              widget=widgets.TextInput(classname="small")),
        Field('type',
              label=u'Тип',
              #conv=convs.EnumChoice(),
              widget=widgets.PopupFilteredSelect()),
        Field('users',
              label=u'Пользователь',
              #conv=convs.ModelChoice(model=models.AdminUser),
              widget=widgets.PopupFilteredSelect()),
        DateFromTo('creation_time'),
        SortField('sort',
                  choices=(('id', 'id'),
                           ('type', 'type'),
                           ('title', 'title'),
                           ('date', 'date')),
                  initial='-id'),
    ]

    def filter_by__streams(self, query, field, value):
        EditLog = self.model
        cond = EditLog.stream_name.in_(value)
        for stream_name in value:
            cond |= EditLog.stream_name.startswith(stream_name + ':')
        return query.filter(cond)

    def filter_by__users(self, query, field, value):
        return query.filter(self.model.users.contains(value))


class EditLogFilterForm(EditLogItemFilterForm):

    fields = [
        Field('streams',
              label=u'Потоки',
              #conv=convs.ListOf(convs.EnumChoice()),
              widget=widgets.PopupFilteredSelect())
    ] + EditLogItemFilterForm.fields





# -*- coding: utf-8 -*-
from iktomi import web
from iktomi.cms.forms import Form, convs
from iktomi.cms.forms.fields import Field, object_ref_fields


class PostNote(web.WebHandler):

    def __init__(self, model):
        self.model = model
        class EditorNoteForm(Form):

            fields = object_ref_fields + [
                Field('body',
                      conv=convs.Char(required=True)),
            ]
        self.EditorNoteForm = EditorNoteForm

    def post_note(self, env, data):
        form = self.EditorNoteForm(env)
        if not form.accept(env.request.POST):
            return env.json({'success': False})
        note = self.model(editor=env.user,
                          **form.python_data)
        env.db.add(note)
        env.db.commit()
        return env.json({'success': True})
    __call__ = post_note




# -*- coding: utf-8 -*-

from iktomi.forms import shortcuts, Field
from iktomi.cms.forms import ModelForm, convs, widgets
from iktomi.cms.stream import ListFields


list_fields = ListFields(('login', u'Логин'),
                         ('name', u'Имя'),
                         ('email', u'email'))
DEFAULT_ROLES = [
    ('wheel',   u'Полный доступ'),
    ('admin',   u'Администратор'),
    ('editor',  u'Редактор'),
]

def clean_password(conv, value):
    form = conv.field.form
    if not value and (not form.item or form.item.id is None):
        raise convs.ValidationError(u'Обязательное поле')
    return value


class RoleConv(convs.EnumChoice):

    @property
    def choices(self):
        return self.field.form.ROLES


class ItemForm(ModelForm):

    ROLES = DEFAULT_ROLES

    fields = [
        Field('name',
              conv=convs.Char(convs.length(3, 250), required=False),
              label=u"Имя"),
        Field('email',
              conv=convs.Email(convs.length(0, 200), required=False),
              label=u"E-mail"),
        Field('login',
              conv=convs.Char(
                    convs.DBUnique(
                        message=u'Объект с таким значением уже существует'),
                    convs.length(3, 250),
                    required=True,
              ),
              label=u"Логин"),
        shortcuts.PasswordSet('password', 
                              filters=(clean_password,),
                              widget=widgets.FieldSetWidget(
                                  template='widgets/fieldset-line'),
                              label=u'пароль',
                              confirm_label=u'подтверждение'),
        Field('active',
              conv=convs.Bool(),
              widget=widgets.CheckBox(),
              initial=True,
              label=u'Активен',
              hint=u'Пользователи с этим флагом могут входить в '
                   u'административный интерфейс'),
        Field('roles', label=u'Роли',
              conv=convs.ListOf(
                  RoleConv()),
              widget=widgets.CheckBoxSelect())
    ]

    def update__password(self, obj, name, value):
        if value is not None:
            obj.set_password(value)

    def update__roles(self, obj, name, value):
        if obj.roles is not None and 'wheel' in obj.roles and \
                'wheel' not in value:
            value.append('wheel')
        if (not obj.roles or 'wheel' not in obj.roles) and 'wheel' in value:
            value.remove('wheel')
        obj.roles = value






# -*- coding: utf-8 -*-
# Common admin views

from webob.exc import HTTPMethodNotAllowed, HTTPForbidden
from iktomi import web
from iktomi.auth import SqlaModelAuth, LoginForm


@web.request_filter
def auth_required(env, data, next_handler):
    if getattr(env, 'user', None) is not None:
        return next_handler(env, data)

    if (env.request.method not in ('GET', 'HEAD') or \
           env.request.is_xhr or \
           '__ajax' in env.request.GET):
        raise HTTPForbidden()
    form = LoginForm(env)
    return env.render_to_response('login.html', dict(form=form))


class AdminAuth(SqlaModelAuth):

    def get_query(self, env, login):
        return SqlaModelAuth.get_query(self, env, login).filter_by(active=True)

    def identify_user(self, env, user_identity):
        user = SqlaModelAuth.identify_user(self, env, user_identity)
        if user is not None and user.active:
            return user

    def login(self):
        def _login(env, data):
            form = self._login_form(env)
            if env.request.method == 'POST':
                if form.accept(env.request.POST):
                    user_identity = self.get_user_identity(
                                                env, **form.python_data)
                    if user_identity is not None:
                        response = env.json({'success': True})
                        return self.login_identity(user_identity,
                                                   response=response)

                return env.json({'success': False, 'errors': form.errors})
            raise HTTPMethodNotAllowed()

        return web.match('/login', 'login') | _login

    def logout(self, redirect_to=None):
        return SqlaModelAuth.logout(self, redirect_to=redirect_to)





# -*- coding: utf-8 -*-
from webob.exc import HTTPMethodNotAllowed
from iktomi.cms.item_lock import ModelLockError, ModelLockedByOther
from iktomi import web
from iktomi.web.shortcuts import Rule


class ItemLockView(object):

    @property
    def app(self):
        return web.method('POST') | web.cases(
            Rule('/_update_lock/<string:item_id>/<string:edit_session>',
                 self.update_lock),
            Rule('/_force_lock/<string:item_id>',
                 self.force_lock),
            Rule('/_release_lock/<string:item_id>/<string:edit_session>',
                 self.release_lock))

    @staticmethod
    def failure_lock_message(e):
        if isinstance(e, ModelLockedByOther):
            return {'status': 'fail',
                    'message': unicode(e),
                    'locked_session': e.edit_session}
        return {'status': 'fail',
                'message': unicode(e)}

    def update_lock(self, env, data):
        if env.request.method != "POST":
            raise HTTPMethodNotAllowed()
        try:
            env.item_lock.update(data.item_id, data.edit_session)
        except ModelLockError, e:
            return env.json(self.failure_lock_message(e))
        return env.json({'status':'updated'})

    def force_lock(self, env, data):
        if env.request.method != "POST":
            raise HTTPMethodNotAllowed()
        try:
            edit_session = env.item_lock.create(data.item_id, True)
        except ModelLockError, e:
            return env.json(self.failure_lock_message(e))

        return env.json({'status':'captured',
                        'edit_session': edit_session})

    def release_lock(self, env, data):
        if env.request.method != "POST":
            raise HTTPMethodNotAllowed()
        env.item_lock.remove(data.item_id, data.edit_session)
        return env.json({'status':'ok'})






# -*- coding: utf-8 -*-
from time import time
from iktomi.cms.item_lock.base import (BaseItemLock, ModelLockError, \
        ModelLockedByOther, ModelLockIsLost, logger)


class MemcachedItemLock(BaseItemLock):

    def create(self, obj, force=False):
        '''Marks model object as editted. Returns edit session on success or
        raises exception. obj should be either model object or its global
        identifier. When force is True the current lock is ignored.'''
        cfg = self.env.cfg
        cache = self.env.cache

        cache.cas_ids.clear()
        edit_session = self._create_edit_session()
        key = self._item_lock_key(obj)
        value = self._item_lock_value(edit_session)
        if force:
            if cache.set(key, value, time=cfg.MODEL_LOCK_TIMEOUT):
                return edit_session
            else:
                raise ModelLockError()
        for i in range(3):
            if cache.add(key, value, time=cfg.MODEL_LOCK_TIMEOUT):
                return edit_session
            old_value = cache.gets(key)
            if old_value is None:
                # Should try add() again
                continue
            if not old_value or \
                    time()-old_value['time']>cfg.MODEL_LOCK_TIMEOUT:
                # Somebody's lock is already expired
                if cache.cas(key, value,
                             time=cfg.MODEL_LOCK_TIMEOUT):
                    return edit_session
                else:
                    continue
            # Somebody holds active lock, no farther attempts
            break
        else:
            logger.error('Failed to lock model object. '\
                         'Problem with memcached?')
            raise ModelLockError()
        lock_user = self.env.db.query(self.env.auth_model)\
                                    .get(old_value['user_id'])
        raise ModelLockedByOther(lock_user, old_value['edit_session'])

    def update(self, obj, edit_session):
        '''Updates model object lock as being active. Raises exception on
        error. obj should be either model object or its global identifier.'''
        cfg = self.env.cfg
        cache = self.env.cache

        cache.cas_ids.clear()
        key = self._item_lock_key(obj)
        for i in range(3):
            old_value = cache.gets(key)
            if not old_value:
                raise ModelLockIsLost()
            elif old_value['edit_session']!=edit_session:
                lock_user = self.env.db.query(self.env.auth_model)\
                                            .get(old_value['user_id'])
                raise ModelLockedByOther(lock_user, old_value['edit_session'])
            new_value = self._item_lock_value(edit_session)
            if cache.cas(key, new_value,
                         time=cfg.MODEL_LOCK_TIMEOUT):
                return
        else:
            # No runtime error here since we want to give user a chance to
            # restore lock.
            raise ModelLockIsLost()

    def remove(self, obj, edit_session):
        '''Removes lock for model object. obj should be either model object or
        its global identifier.'''
        cache = self.env.cache

        cache.cas_ids.clear()
        key = self._item_lock_key(obj)
        # We can't garuantee memcache's delete method will remove only our
        # lock, so we update the record with empty value and minimal (1 sec)
        # timeout.
        old_value = cache.gets(key)
        # It's too late to do something in case of error, so we just ignore
        # returned value.
        if old_value and old_value['edit_session']==edit_session:
            cache.cas(key, '', time=1)

    def check(self, obj):
        cache = self.env.cache
        key = self._item_lock_key(obj)
        value = cache.get(key)
        return value or None




# -*- coding: utf-8 -*-
from webob.exc import HTTPNotFound
from functools import partial
from iktomi import web
from iktomi.utils.storage import VersionedStorage
from iktomi.cms.stream import decode_stream_uid
from iktomi.cms.stream_actions import GetAction
from iktomi.cms.loner import Loner
from iktomi.utils.paginator import ModelPaginator, FancyPageRange
from .forms import EditLogFilterForm, EditLogItemFilterForm


def _preload_users(env, entries):
    entries_by_id = dict((x.id, x) for x in entries)
    m = env.models_.admin
    query = env.db.query(m.AdminUser)\
                  .join(m.EditLogAdminUser)\
                  .filter(m.EditLogAdminUser.log_id.in_(entries_by_id))\
                  .with_entities(m.EditLogAdminUser.log_id, m.AdminUser)
    objects = query.all()

    for entry in entries_by_id.values():
        # XXX what is the right way to do this?!
        entry.__dict__['users'] = [v for k, v in objects if k == entry.id]


def _preload_items(env, entries):
    by_stream = {}
    for entry in entries:
        by_stream.setdefault(entry.stream_name, [])
        by_stream[entry.stream_name].append(entry.object_id)

    by_stream_and_id = {}
    for stream_name, ids in by_stream.items():

        name = stream_name.split(':')[0]
        if name not in env.streams:
            continue
        stream = env.streams[name]
        lang = decode_stream_uid(stream_name)[1].get('lang')

        stream_env = VersionedStorage(version="admin", lang=lang)
        stream_env._storage._parent_storage = env
        # XXX only for multi db! will fail on single db
        stream_env.models = env.models_.admin
        if lang:
            stream_env.models = getattr(stream_env.models, lang)

        M = stream.get_model(stream_env)
        items = stream.item_query(stream_env).filter(M.id.in_(ids)).all()
        for item in items:
            by_stream_and_id[(stream_name, str(item.id))] = item
    return by_stream_and_id


def _expand(env, obj, for_item=None):
    # XXX all this is like a hack!
    stream_name = obj.stream_name.split(':')[0]
    if stream_name not in env.streams:
        return {}
    stream = env.streams[stream_name]
    lang = decode_stream_uid(obj.stream_name)[1].get('lang')
    if for_item is None:
        stream_env = VersionedStorage(version="admin", lang=lang)
        stream_env._storage._parent_storage = env
        # XXX only for multi db! will fail on single db
        stream_env.models = env.models.admin
        if lang:
            stream_env.models = getattr(stream_env.models, lang)
    else:
        stream_env = env

    assert 'r' in stream.get_permissions(stream_env)
    if for_item is None:
        item = stream.item_query(stream_env).filter_by(id=obj.object_id).first()
    else:
        item = for_item

    if item is not None:
        item_title = getattr(item, 'title', unicode(item))
    else:
        item_title = u"<{}: удалённый или недоступный объект: {}>".format(
                                stream.title, obj.object_id)

    return {"obj": obj,
            "stream": stream,
            "item": item,
            "item_title": item_title,
            "lang": lang,
            "type": stream.edit_log_action.log_type_title(obj)}


def global_log(env, data, stream=None):
    #insure_is_xhr(env)

    EditLog = env.edit_log_model

    assert (stream and getattr(data, 'item', None)) or \
           not (stream or getattr(data, 'item', None))

    if stream is None:
        query = env.db.query(EditLog)

        stream_names = [k for (k, s) in env.streams.items()
                        if 'r' in s.get_permissions(env) and s.edit_log_action is not None]
        stream_choices = [(k, env.streams[k].title)
                          for k in stream_names]
        types = {}
        for k in stream_names:
            types.update(env.streams[k].edit_log_action.log_types)
        types = types.items()

        filter_form = EditLogFilterForm(env,
                                        streams=stream_choices,
                                        models=env.models_.admin,
                                        types=types)
    else:
        log_types = stream.edit_log_action.log_types
        filter_form = EditLogItemFilterForm(env,
                                            models=env.models,
                                            types=log_types.items())
        query = EditLog.query_for_item(env.db, data.item)

    filter_form.model = EditLog
    filter_form.accept(env.request.GET)
    query = filter_form.filter(query)

    #url = env.root.build_subreverse(env.current_location)
    #url = stream.url_for(env, self.action, item=data.item.id)
    paginator = ModelPaginator(env.request, query,
                               impl=FancyPageRange(),
                               limit=50,
                               #url=url
                               )
    _preload_users(env, paginator.items)

    if getattr(data, 'item', None) is None:
        expand_obj = partial(_expand, env)
        by_stream_and_id = _preload_items(env, paginator.items)
        paginator.items = [expand_obj(obj,
                                      for_item=by_stream_and_id.get((obj.stream_name, obj.object_id)))
                           for obj in paginator.items]

    else:
        expand_obj = partial(_expand, env, for_item=data.item)
        paginator.items = [expand_obj(obj) for obj in paginator.items]

    paginator.items = filter(None, paginator.items)

    reverse_data = data.as_dict()
    if stream is not None:
        # XXX hack!
        reverse_data = dict(reverse_data,
                            item=data.item.id)
    current_url = env.root.build_subreverse(env.current_location, **reverse_data)

    tdata = dict(paginator=paginator,
                 title=u"Журнал изменений",
                 filter_form=filter_form,
                 current_url=current_url,
                 no_layout='__no_layout' in env.request.GET,
                 )
    if stream is not None:
        tdata = dict(tdata, item=data.item,
                            stream=stream)

    return env.render_to_response('edit_log/index', tdata)


class EditLogHandler(GetAction):

    item_lock = False
    allowed_for_new = False
    display = False
    cls = 'edit-log'
    action = 'edit_log'
    title = u'Журнал изменений'
    log_types = {
        'create': u'Создано',
        'edit': u'Правка',
        'publish': u'Опубликовано',
        'delete': u'Удалено',
        'unpublish': u'Снято с публикации',
        'revert': u'Откат'
    }

    @property
    def PrepareItemHandler(self):
        return self.stream.edit_action.PrepareItemHandler

    @property
    def app(self):
        if isinstance(self.stream, Loner):
            prefix = web.prefix('/log', name=self.action)
        else:
            prefix = web.prefix('/<idconv:item>/log', name=self.action,
                                convs={'idconv': self.stream.id_converter})
        return prefix | self.PrepareItemHandler(self) | web.cases(
                web.match('') | self,
                web.match('/<int:log_id>', 'entry') | self.log_entry,
                )

    @classmethod
    def log_type_title(cls, log):
        return cls.log_types[log.type]

    def edit_log(self, env, data):
        return global_log(env, data, stream=self.stream)
    __call__ = edit_log

    def log_entry(self, env, data):
        #insure_is_xhr(env)

        if getattr(env, 'version', None) == 'front':
            raise HTTPNotFound()

        EditLog = env.edit_log_model
        log = EditLog.query_for_item(env.db, data.item)\
                     .filter_by(id=data.log_id)\
                     .first()

        if log is None:
            raise HTTPNotFound()

        rel_stream_name, params = decode_stream_uid(log.stream_name)
        if rel_stream_name in env.streams:
            rel_stream = env.streams[rel_stream_name]
        else:
            # Deleted or renamed stream
            raise NotImplementedError("Deleted or renamed stream", rel_stream_name)
        rel_env = VersionedStorage(**params)
        rel_env._storage._parent_storage = env

        form_cls = rel_stream.config.ItemForm

        item_cp = type(data.item)(id=data.item.id)

        form1 = form_cls.load_initial(env, item_cp, initial={})
        form2 = form_cls.load_initial(env, item_cp, initial={})
        form1.model = form2.model = rel_stream.get_model(env)

        form1.accept(log.before)
        form2.accept(log.after)

        # reset permissions
        form1.permissions = form2.permissions = frozenset('r')
        form1.fields = [field(parent=form1) for field in form1.fields]
        form2.fields = [field(parent=form2) for field in form2.fields]

        diff = form1.get_diff(form2)
        diffs = diff['children'] if diff is not None else []

        return env.render_to_response('edit_log/item',
                                      dict(form1=form1,
                                           form2=form2,
                                           diffs=diffs,
                                           log=log,
                                           stream=self.stream,
                                           item=data.item,
                                           log_type=self.log_type_title(log)))





import json
import requests
from collections import OrderedDict
from sys import argv
import sys

def open_file(fname):
    return json.loads(open(fname).read())

def fetch_descriptions(lang, chars):
    chars = sorted(u''.join([k+v for k, v in chars.items()]))


    catalogs = {}
    result = OrderedDict()
    base_url = "https://raw.githubusercontent.com/unicode-table/unicode-table-data/master/loc/{}/symbols/{}.txt"

    for char in chars:
        o = ord(char)
        catalog_name = "%0.4X" % (o - (o % 256))
        char_code = "%0.4X" % o
        if not catalog_name in catalogs:
            url = base_url.format(lang, catalog_name)
            catalog = requests.get(url).content.splitlines()
            catalog = [l.split(':') for l in catalog if l.strip() and 
                                                        not l.strip().startswith('#') and
                                                        ':' in l]
            catalog = [(l[0].strip(), l[1].strip()) for l in catalog]
            catalog = dict(catalog)
            catalogs[catalog_name] = catalog
        else:
            catalog = catalogs[catalog_name]

        result[char] = catalog.get(char_code, "")
    return result

def print_json(result):
    raw_json = json.dumps(result, indent=1, ensure_ascii=False, encoding='utf8')
    print(raw_json.encode('utf8'))

if __name__ == '__main__':
    if len(argv) != 3 or len(argv[1]) != 2:
        sys.stderr.write("Usage:\n\n    load_unicode_descriptions.py en chars.json\n\n")
        exit(1)
    chars = open_file(argv[2])
    result = fetch_descriptions(argv[1], chars)
    print_json(result)





# -*- coding: utf-8 -*-
from iktomi import web
from iktomi.cms.loner import Loner
from iktomi.utils import cached_property
from .stream import PublishItemHandler, PublishAction, UnpublishAction, \
            RevertAction, PublishStreamNoState
from ..loner import LonerHandler, PrepareLonerHandler


class PublishLonerHandler(LonerHandler, PublishItemHandler):
    pass


class BaseLonerAction(object):
    PrepareItemHandler = PrepareLonerHandler

    @property
    def app(self):
        return web.match('/%s' % self.action, self.action) | \
               web.method('POST', strict=True) | \
               self.PrepareItemHandler(self) | self


class LonerPublishAction(BaseLonerAction, PublishAction):
    pass


class LonerUnpublishAction(BaseLonerAction, UnpublishAction):
    pass


class LonerRevertAction(BaseLonerAction, RevertAction):
    pass


class PublishLoner(PublishStreamNoState, Loner):

    core_actions = [PublishLonerHandler(),
                    LonerPublishAction(),
                    LonerRevertAction(),
                    LonerUnpublishAction()]

    def url_for(self, env, name=None, **kwargs):
        kwargs.setdefault('version', getattr(env, 'version', self.versions[0][0]))
        return Loner.url_for(self, env, name, **kwargs)

    @cached_property
    def perms(self):
        p = getattr(self.config, 'permissions', {})
        p.setdefault('wheel', 'rwp')
        return p


class PublishLonerNoState(PublishLoner):

    core_actions = [PublishLonerHandler(),
                    LonerPublishAction(),
                    LonerRevertAction()]



